import logging

import pytest

from ini2toml.plugins import profile_independent_tasks as tasks
from ini2toml.plugins.setuptools_pep621 import SetuptoolsPEP621, activate, isdirective
from ini2toml.translator import Translator


@pytest.fixture
def plugin():
    return SetuptoolsPEP621()


@pytest.fixture
def translator(plugin):
    return Translator(plugins=[activate])


@pytest.fixture
def parse(translator):
    return lambda text: translator.loads(text)


@pytest.fixture
def convert(translator):
    return lambda irepr: translator.dumps(irepr)


example_normalise_keys = """\
[metadata]
summary = Automatically translates .cfg/.ini files into TOML
author_email = example@example
license-file = LICENSE.txt
long_description_content_type = text/x-rst; charset=UTF-8
home_page = https://github.com/abravalheri/ini2toml/
classifier = Development Status :: 4 - Beta
platform = any
"""
expected_normalise_keys = """\
[metadata]
description = "Automatically translates .cfg/.ini files into TOML"
author-email = "example@example"
license-files = "LICENSE.txt"
long-description-content-type = "text/x-rst; charset=UTF-8"
url = "https://github.com/abravalheri/ini2toml/"
classifiers = "Development Status :: 4 - Beta"
platforms = "any"
"""


def test_normalise_keys(plugin, parse, convert):
    obj = parse(example_normalise_keys)
    obj = plugin.normalise_keys(obj)
    assert convert(obj) == expected_normalise_keys


# ----


example_convert_directives = """\
[metadata]
version = attr: mymodule.myfunc
classifiers = file: CLASSIFIERS.txt
description = file: README.txt

[options]
entry-points = file: ENTRYPOINTS.txt
packages = find_namespace:
"""
expected_convert_directives = """\
[metadata]
version = {attr = "mymodule.myfunc"}
classifiers = {file = "CLASSIFIERS.txt"}
description = {file = "README.txt"}

[options]
entry-points = {file = "ENTRYPOINTS.txt"}
packages = {find_namespace = ""}
"""


def test_convert_directives(plugin, parse, convert):
    doc = parse(example_convert_directives)
    print(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    doc = plugin.convert_directives(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    print(doc)
    assert convert(doc) == expected_convert_directives


# ----


example_apply_value_processing = """\
[metadata]
version = 5.3  # comment
classifiers =
    Development Status :: 4 - Beta
    Programming Language :: Python
keywords = python, module
[options]
zip-safe = False  # comment
package-dir =
    *=src # TODO: tomlkit/atoml bug with empty keys
install-requires =
    importlib-metadata; python_version<"3.8"
    configupdater>=3,<=4
[options.entry-points]
# For example:
console-scripts =
    putup = pyscaffold.cli:run  # CLI exec
pyscaffold.cli =
    # comment
    fibonacci = ini2toml.skeleton:run # comment
    awesome = pyscaffoldext.awesome.extension:AwesomeExtension
"""
expected_apply_value_processing = """\
[metadata]
version = "5.3" # comment
classifiers = [
    "Development Status :: 4 - Beta",
    "Programming Language :: Python",
]
keywords = ["python", "module"]

[options]
zip-safe = false # comment
package-dir = {"*" = "src"} # TODO: tomlkit/atoml bug with empty keys
install-requires = [
    "importlib-metadata; python_version<\\"3.8\\"",
    "configupdater>=3,<=4",
]

["project:entry-points"]
# For example:

["project:entry-points"."pyscaffold.cli"]
# comment
fibonacci = "ini2toml.skeleton:run" # comment
awesome = "pyscaffoldext.awesome.extension:AwesomeExtension"

["project:scripts"]
putup = "pyscaffold.cli:run" # CLI exec
"""


def test_move_entry_points_and_apply_value_processing(plugin, parse, convert):
    doc = parse(example_apply_value_processing)
    print(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    doc = plugin.move_and_split_entrypoints(doc)
    doc = plugin.apply_value_processing(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    print(doc)
    text = tasks.remove_trailing_spaces(convert(doc)).strip()
    assert text == expected_apply_value_processing.strip()


# ----

example_separate_subtables = """\
[options.packages.find]
where = src
[project:entry-points]
# For example:
[project:scripts]
django-admin = django.core.management:execute_from_command_line
"""

expected_separate_subtables = """\
[tool]
[tool.setuptools]
[tool.setuptools.packages]
[tool.setuptools.packages.find]
where = "src"

[project]
[project.entry-points]
# For example:

[project.scripts]
django-admin = "django.core.management:execute_from_command_line"
"""


def test_split_subtables(plugin, parse, convert):
    doc = parse(example_separate_subtables.strip())
    print(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    doc = plugin.split_subtables(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    print(doc)
    assert convert(doc).strip() == expected_separate_subtables.strip()


# ----

example_entrypoints_and_split_subtables = """\
[options.packages.find]
where = src
[options.entry-points]
# For example:
console-scripts =
    django-admin = django.core.management:execute_from_command_line
gui-scripts =
    project = my.module:function [extra-dep]
pyscaffold.cli =
    config = pyscaffold.extensions.config:Config
    interactive = pyscaffold.extensions.interactive:Interactive
"""

expected_entrypoints_and_split_subtables = """\
[tool]
[tool.setuptools]
[tool.setuptools.packages]
[tool.setuptools.packages.find]
where = "src"

[project]
[project.entry-points]
# For example:

[project.entry-points."pyscaffold.cli"]
config = "pyscaffold.extensions.config:Config"
interactive = "pyscaffold.extensions.interactive:Interactive"

[project.scripts]
django-admin = "django.core.management:execute_from_command_line"

[project.gui-scripts]
project = "my.module:function [extra-dep]"
"""


def test_entrypoints_and_split_subtables(plugin, parse, convert):
    doc = parse(example_entrypoints_and_split_subtables.strip())
    print(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    doc = plugin.move_and_split_entrypoints(doc)
    doc = plugin.apply_value_processing(doc)
    doc = plugin.split_subtables(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    print(doc)
    assert convert(doc).strip() == expected_entrypoints_and_split_subtables.strip()


# ----


example_fix_license = """\
[metadata]
license = MPL-2.0
license-files = LICENSE.txt
"""

expected_fix_license = """\
[metadata]
[metadata.license]
file = "LICENSE.txt"
"""


def test_merge_license_and_files(plugin, parse, convert):
    doc = parse(example_fix_license.strip())
    print(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    doc = plugin.merge_license_and_files(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    print(doc)
    assert convert(doc).strip() == expected_fix_license.strip()


# ----


example_fix_packages = """\
[options]
packages = find-namespace:
[options.packages.find]
where = src
exclude =
    tests
"""

expected_fix_packages = """\
[options]

["options.packages.find-namespace"]
where = "src"
exclude = ["tests"]
"""


def test_fix_packages(plugin, parse, convert):
    doc = parse(example_fix_packages.strip())
    print(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    doc = plugin.fix_packages(doc)
    doc = plugin.apply_value_processing(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    print(doc)
    assert convert(doc).strip() == expected_fix_packages.strip()


# ----


example_fix_setup_requires = """\
[options]
setup-requires =
    setuptools>=46.1.0
    setuptools_scm>=5
"""

expected_fix_setup_requires = """\
[build-system]
requires = [
    "setuptools>=46.1.0",
    "setuptools_scm>=5",
    "wheel",
]
build-backend = "setuptools.build_meta"
"""


def test_fix_setup_requires(plugin, parse, convert):
    doc = plugin.template()
    doc.update(parse(example_fix_setup_requires.strip()))
    print(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    doc = plugin.fix_setup_requires(doc)
    doc = plugin.apply_value_processing(doc)
    doc = plugin.move_setup_requires(doc)
    doc.pop("tool", None)
    doc.pop("options", None)
    doc.pop("metadata", None)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    print(doc)
    text = tasks.remove_trailing_spaces(convert(doc)).strip()
    assert text == expected_fix_setup_requires.strip()


example_dynamic = """\
[metadata]
version = attr: django.__version__
classifiers = file: classifiers.txt
description = file: readme.txt

[options]
entry-points = file: entry-points.txt
"""

expected_dynamic = """\
[metadata]

dynamic = [
    "version",
    "classifiers",
    "description",
    "entry-points",
    "scripts",
    "gui-scripts",
]

["options.dynamic"]
version = {attr = "django.__version__"}
classifiers = {file = "classifiers.txt"}
description = {file = "readme.txt"}
entry-points = {file = "entry-points.txt"}
"""  # noqa


def test_isdirective(plugin, parse, convert):
    doc = parse(example_dynamic.strip())
    assert isdirective(doc["metadata"]["version"], ("attr",))
    assert isdirective(doc["metadata"]["classifiers"])
    assert isdirective(doc["metadata"]["description"], ("file",))
    assert isdirective(doc["options"]["entry-points"])
    assert not isdirective("")
    assert not isdirective("some value")


def test_fix_dynamic(plugin, parse, convert):
    doc = parse(example_dynamic.strip())
    print(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    doc = plugin.fix_dynamic(doc)
    doc.pop("tool", None)
    doc.pop("options", None)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    print(doc)
    assert convert(doc).strip() == expected_dynamic.strip()


# ----

expected_empty = """\
[project]
dynamic = ["version"]

[build-system]
requires = ["setuptools", "wheel"]
build-backend = "setuptools.build_meta"

[tool]
[tool.setuptools]
"""


def test_empty(translator, plugin, parse, convert):
    doc = parse("")
    print(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    doc = plugin.normalise_keys(doc)
    doc = plugin.pep621_transform(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    print(doc)
    assert convert(doc).strip() == expected_empty.strip()

    # Same thing but with the higher level API:
    text = translator.translate("", profile_name="setup.cfg")
    assert text.strip() == expected_empty.strip()


# ----
example_data_files = """
[options]
data-files =
    a = b
"""

expected_data_files = """\
[project]
dynamic = ["version"]

[build-system]
requires = ["setuptools", "wheel"]
build-backend = "setuptools.build_meta"

[tool]
[tool.setuptools]
data-files = {a = ["b"]}
"""


def test_data_files(translator, caplog):
    # Same thing but with the higher level API:
    with caplog.at_level(logging.DEBUG):
        text = translator.translate(example_data_files, profile_name="setup.cfg")
        assert text.strip() == expected_data_files.strip()

    assert "'data-files' is deprecated" in caplog.text



from dataclasses import dataclass, field
from dataclasses import replace as _replace
from typing import List, Optional, TypeVar

from .types import IntermediateProcessor, ProfileAugmentationFn, TextProcessor

P = TypeVar("P", bound="Profile")


@dataclass
class Profile:
    """Profile object that follows the public API defined in
    :class:`ini2toml.types.Profile`.
    """

    name: str
    help_text: str = ""
    pre_processors: List[TextProcessor] = field(default_factory=list)
    intermediate_processors: List[IntermediateProcessor] = field(default_factory=list)
    post_processors: List[TextProcessor] = field(default_factory=list)
    ini_parser_opts: Optional[dict] = None

    def replace(self: P, **changes) -> P:
        """See :func:`dataclasses.replace`"""
        return _replace(self, **changes)


@dataclass
class ProfileAugmentation:
    fn: ProfileAugmentationFn
    active_by_default: bool = False
    name: str = ""
    help_text: str = ""

    def is_active(self, explicitly_active: Optional[bool] = None) -> bool:
        """``explicitly_active`` is a tree-state variable: ``True`` if the user
        explicitly asked for the augmentation, ``False`` if the user explicitly denied
        the augmentation, or ``None`` otherwise.
        """
        activation = explicitly_active
        return activation is True or (activation is None and self.active_by_default)



from textwrap import dedent

import tomli

from ini2toml.drivers import full_toml, lite_toml
from ini2toml.plugins import mypy
from ini2toml.translator import Translator


def test_mypy():
    example = """\
    [mypy]
    python_version = 2.7
    warn_return_any = True
    warn_unused_configs = True
    plugins = mypy_django_plugin.main, returns.contrib.mypy.returns_plugin
    [mypy-mycode.foo.*]
    disallow_untyped_defs = True
    [mypy-mycode.bar]
    warn_return_any = False
    [mypy-somelibrary,some_other_library]
    ignore_missing_imports = True
    """
    expected = """\
    [mypy]
    python_version = "2.7"
    warn_return_any = true
    warn_unused_configs = true
    plugins = ["mypy_django_plugin.main", "returns.contrib.mypy.returns_plugin"]

    [[mypy.overrides]]
    module = ["mycode.foo.*"]
    disallow_untyped_defs = true

    [[mypy.overrides]]
    module = ["mycode.bar"]
    warn_return_any = false

    [[mypy.overrides]]
    module = ["somelibrary", "some_other_library"]
    ignore_missing_imports = true
    """
    for convert in (lite_toml.convert, full_toml.convert):
        translator = Translator(plugins=[mypy.activate], toml_dumps_fn=convert)
        out = translator.translate(dedent(example), "mypy.ini").strip()
        expected = dedent(expected).strip()
        print("expected=\n" + expected + "\n***")
        print("out=\n" + out)
        try:
            assert expected == out
        except AssertionError:
            # At least the Python-equivalents when parsing should be the same
            assert tomli.loads(expected) == tomli.loads(out)



from textwrap import dedent

import tomli

from ini2toml.drivers import full_toml, lite_toml
from ini2toml.plugins import coverage
from ini2toml.translator import Translator


def test_coverage():
    example = """\
    # .coveragerc to control coverage.py
    [run]
    branch = True
    source = ini2toml
    # omit = bad_file.py
    [paths]
    source =
        src/
        */site-packages/
    [report]
    # Regexes for lines to exclude from consideration
    exclude_lines =
        # Have to re-enable the standard pragma
        pragma: no cover
        # Don't complain about missing debug-only code
        def __repr__
        # Don't complain if tests don't hit defensive assertion code
        raise AssertionError
        raise NotImplementedError
        # Don't complain if non-runnable code isn't run
        if 0:
        if __name__ == .__main__.:
    """
    expected = """\
    # .coveragerc to control coverage.py

    [run]
    branch = true
    source = ["ini2toml"]
    # omit = bad_file.py

    [paths]
    source = [
        "src/", 
        "*/site-packages/", 
    ]

    [report]
    # Regexes for lines to exclude from consideration
    exclude_lines = [
        # Have to re-enable the standard pragma
        "pragma: no cover", 
        # Don't complain about missing debug-only code
        "def __repr__", 
        # Don't complain if tests don't hit defensive assertion code
        "raise AssertionError", 
        "raise NotImplementedError", 
        # Don't complain if non-runnable code isn't run
        "if 0:", 
        "if __name__ == .__main__.:", 
    ]
    """
    for convert in (full_toml.convert, lite_toml.convert):
        translator = Translator(plugins=[coverage.activate], toml_dumps_fn=convert)
        out = translator.translate(dedent(example), ".coveragerc").strip()
        expected = dedent(expected).strip()
        print("expected=\n" + expected + "\n***")
        print("out=\n" + out)
        try:
            assert expected == out
        except AssertionError:
            # At least the Python-equivalents when parsing should be the same
            assert tomli.loads(expected) == tomli.loads(out)



# The code in this file is mostly borrowed/adapted from PyScaffold and was originally
# published under the MIT license.
# The original PyScaffold license can be found in 'tests/examples/pyscaffold'

import sys

import pytest

from ini2toml import plugins
from ini2toml.plugins import ENTRYPOINT_GROUP, ErrorLoadingPlugin

EXISTING = (
    "setuptools_pep621",
    "best_effort",
    "isort",
    "coverage",
    "pytest",
    "mypy",
    "independent_tasks",
)


if sys.version_info[:2] >= (3, 8):
    # TODO: Import directly (no need for conditional) when `python_requires = >= 3.8`
    from importlib.metadata import EntryPoint  # pragma: no cover
else:
    from importlib_metadata import EntryPoint  # pragma: no cover


def test_load_from_entry_point__error():
    # This module does not exist, so Python will have some trouble loading it
    # EntryPoint(name, value, group)
    entry = "mypkg.SOOOOO___fake___:activate"
    fake = EntryPoint("fake", entry, ENTRYPOINT_GROUP)
    with pytest.raises(ErrorLoadingPlugin):
        plugins.load_from_entry_point(fake)


def test_iterate_entry_points():
    plugin_iter = plugins.iterate_entry_points()
    assert hasattr(plugin_iter, "__iter__")
    pluging_list = list(plugin_iter)
    assert all([isinstance(e, EntryPoint) for e in pluging_list])
    name_list = [e.name for e in pluging_list]
    for ext in EXISTING:
        assert ext in name_list


def test_list_from_entry_points():
    # Should return a list with all the plugins registered in the entrypoints
    pluging_list = plugins.list_from_entry_points()
    orig_len = len(pluging_list)
    assert all(callable(e) for e in pluging_list)
    plugin_names = " ".join(str(e.__module__) for e in pluging_list)
    for example in EXISTING:
        assert example in plugin_names

    # a filtering function can be passed to avoid loading plugins that are not needed
    pluging_list = plugins.list_from_entry_points(filtering=lambda e: e.name != "isort")
    plugin_names = " ".join(str(e.__module__) for e in pluging_list)
    assert len(pluging_list) == orig_len - 1
    assert "isort" not in plugin_names



from textwrap import dedent

import tomli

from ini2toml.drivers import full_toml, lite_toml
from ini2toml.plugins import isort
from ini2toml.translator import Translator


def test_isort():
    example = """\
    [{section}]
    profile = black
    order_by_type = false
    src_paths=isort,test
    known_first_party = ini2toml
    combine_as_imports = true
    default_section = THIRDPARTY
    include_trailing_comma = true
    line_length = 79
    multi_line_output = 5
    """
    expected_template = """\
    [{section}]
    profile = "black"
    order_by_type = false
    src_paths = ["isort", "test"]
    known_first_party = ["ini2toml"]
    combine_as_imports = true
    default_section = "THIRDPARTY"
    include_trailing_comma = true
    line_length = 79
    multi_line_output = 5
    """
    for convert in (lite_toml.convert, full_toml.convert):
        translator = Translator(plugins=[isort.activate], toml_dumps_fn=convert)
        for file, section in [(".isort.cfg", "settings"), ("setup.cfg", "isort")]:
            expected = dedent(expected_template.format(section=section)).strip()
            out = translator.translate(dedent(example).format(section=section), file)
            print("expected=\n" + expected + "\n***")
            print("out=\n" + out)
            try:
                assert expected in out
            except AssertionError:
                # At least the Python-equivalents when parsing should be the same
                assert tomli.loads(expected) == tomli.loads(out)



from functools import reduce
from types import MappingProxyType
from typing import Dict, Generic, List, Mapping, Sequence, TypeVar

from . import types  # Structural/Abstract types
from .errors import (
    AlreadyRegisteredAugmentation,
    InvalidAugmentationName,
    UndefinedProfile,
)
from .profile import Profile, ProfileAugmentation
from .transformations import apply

T = TypeVar("T")
EMPTY = MappingProxyType({})  # type: ignore


class BaseTranslator(Generic[T]):
    """Translator object that follows the public API defined in
    :class:`ini2toml.types.Translator`. See :doc:`dev-guide` for a quick explanation of
    concepts such as plugins, profiles, profile augmentations, etc.

    Arguments
    ---------

    ini_loads_fn:
        function to convert the ``.ini/.cfg`` file into an :class:`intermediate
        representation <ini2toml.intermediate_repr.IntermediateRepr>` object.
        Possible values for this argument include:

        - :func:`ini2toml.drivers.configparser.parse` (when comments can be simply
          removed)
        - :func:`ini2toml.drivers.configupdater.parse` (when you wish to preserve
          comments in the TOML output)

    toml_dumps_fn:
        function to convert the :class:`intermediate representation
        <ini2toml.intermediate_repr.IntermediateRepr>` object into (ideally)
        a TOML string.
        If you don't exactly need a TOML string (maybe you want your TOML to
        be represented by :class:`bytes` or simply the equivalent :obj:`dict`) you can
        also pass a ``Callable[[IntermediateRepr], T]`` function for any desired ``T``.

        Possible values for this argument include:

        - :func:`ini2toml.drivers.lite_toml.convert` (when comments can be simply
          removed)
        - :func:`ini2toml.drivers.full_toml.convert` (when you wish to preserve
          comments in the TOML output)
        - :func:`ini2toml.drivers.plain_builtins.convert` (when you wish to retrieve a
          :class:`dict` equivalent to the TOML, instead of string with the TOML syntax)

    plugins:
        list of plugins activation functions. By default no plugin will be activated.
    profiles:
        list of profile objects, by default no profile will be pre-loaded (plugins can
        still add them).
    profile_augmentations:
        list of profile augmentations. By default no profile augmentation will be
        preloaded (plugins can still add them)
    ini_parser_opts:
        syntax options for parsing ``.ini/.cfg`` files (see
        :mod:`~configparser.ConfigParser` and :mod:`~configupdater.ConfigUpdater`).
        By default it uses the standard configuration of the selected parser (depending
        on the choice of ``ini_loads_fn``).

    Tip
    ---

    Most of the times the usage of :class:`~ini2toml.translator.Translator` is preferred
    over :class:`~ini2toml.base_translator.BaseTranslator` (unless you are vendoring
    ``ini2toml`` and wants to reduce the number of files included in your project).
    """

    profiles: Dict[str, types.Profile]
    plugins: List[types.Plugin]

    def __init__(
        self,
        ini_loads_fn: types.IniLoadsFn,
        toml_dumps_fn: types.IReprCollapseFn[T],
        plugins: Sequence[types.Plugin] = (),
        profiles: Sequence[types.Profile] = (),
        profile_augmentations: Sequence[types.ProfileAugmentation] = (),
        ini_parser_opts: Mapping = EMPTY,
    ):
        self.plugins = list(plugins)
        self.ini_parser_opts = ini_parser_opts
        self.profiles = {p.name: p for p in profiles}
        self.augmentations: Dict[str, types.ProfileAugmentation] = {
            (p.name or p.fn.__name__): p for p in profile_augmentations
        }

        self._loads_fn = ini_loads_fn
        self._dumps_fn = toml_dumps_fn

        for activate in self.plugins:
            activate(self)

    def loads(self, text: str) -> types.IntermediateRepr:
        return self._loads_fn(text, self.ini_parser_opts)

    def dumps(self, irepr: types.IntermediateRepr) -> T:
        return self._dumps_fn(irepr)

    def __getitem__(self, profile_name: str) -> types.Profile:
        """Retrieve an existing profile (or create a new one)."""
        if profile_name not in self.profiles:
            profile = Profile(profile_name)
            if self.ini_parser_opts:
                profile = profile.replace(ini_parser_opts=self.ini_parser_opts)
            self.profiles[profile_name] = profile
        return self.profiles[profile_name]

    def augment_profiles(
        self,
        fn: types.ProfileAugmentationFn,
        active_by_default: bool = False,
        name: str = "",
        help_text: str = "",
    ):
        """Register a profile augmentation function to be called after the
        profile is selected and before the actual translation (see :doc:`dev-guide`).
        """
        name = (name or fn.__name__).strip()
        InvalidAugmentationName.check(name)
        AlreadyRegisteredAugmentation.check(name, fn, self.augmentations)
        help_text = help_text or fn.__doc__ or ""
        obj = ProfileAugmentation(fn, active_by_default, name, help_text)
        self.augmentations[name] = obj

    def _add_augmentations(
        self, profile: types.Profile, explicit_activation: Mapping[str, bool] = EMPTY
    ) -> types.Profile:
        for aug in self.augmentations.values():
            if aug.is_active(explicit_activation.get(aug.name)):
                aug.fn(profile)
        return profile

    def translate(
        self,
        ini: str,
        profile_name: str,
        active_augmentations: Mapping[str, bool] = EMPTY,
    ) -> T:
        UndefinedProfile.check(profile_name, list(self.profiles.keys()))
        profile = self._add_augmentations(self[profile_name], active_augmentations)

        ini = reduce(apply, profile.pre_processors, ini)
        irepr = self.loads(ini)
        irepr = reduce(apply, profile.intermediate_processors, irepr)
        toml = self.dumps(irepr)
        return reduce(apply, profile.post_processors, toml)



import pytest

from ini2toml.intermediate_repr import Commented, CommentedKV
from ini2toml.intermediate_repr import IntermediateRepr as IR


class TestInteremediateRepr:
    def test_init(self):
        with pytest.raises(ValueError):
            # elements and order must have the same elements
            IR({"a": 1, "b": 2}, ["a", "c"])

    def test_eq(self):
        assert IR({"a": 1}, inline_comment="hello world") != IR({"a": 1})

    def test_rename(self):
        irepr = IR({"a": 1, "b": 2})
        with pytest.raises(KeyError):
            # Cannot rename to an existing key
            irepr.rename("a", "b")

        irepr.rename("a", "c")
        assert irepr == IR({"c": 1, "b": 2})

        with pytest.raises(KeyError):
            # Cannot rename to a missing key unless explicitly ignoring
            irepr.rename("a", "b")

        irepr.rename("a", "d", ignore_missing=True)
        assert irepr == IR({"c": 1, "b": 2})

    def test_insert(self):
        irepr = IR({"a": 1, "b": 2})
        with pytest.raises(KeyError):
            # Cannot insert an existing key
            irepr.insert(0, "a", 3)

    def test_copy(self):
        irepr = IR({"a": 1, "b": 2})
        other = irepr.copy()
        other["a"] = 3
        assert irepr != other
        assert other["a"] == 3
        assert irepr["a"] == 1


class TestCommentedKV:
    def test_find(self):
        v = CommentedKV([Commented([("a", 1), ("b", 2)]), Commented([("c", 3)])])
        assert v.find("a") == (0, 0)
        assert v.find("b") == (0, 1)
        assert v.find("c") == (1, 0)
        assert v.find("d") is None



"""This module serves as a compatibility layer between API-compatible
style preserving TOML editing libraries (e.g. atoml and atoml).
It makes it easy to swap between implementations for testing (by means of search and
replace).
"""
from collections.abc import Mapping, MutableSequence, Sequence
from functools import singledispatch
from typing import Iterable, Optional, Tuple, TypeVar, Union, cast

from atoml import (
    aot,
    array,
    comment,
    document,
    dumps,
    inline_table,
    item,
    loads,
    nl,
    table,
)
from atoml.items import AoT, Array, InlineTable, Item, Table, Whitespace
from atoml.toml_document import TOMLDocument

from ..errors import InvalidTOMLKey
from ..types import (
    KV,
    Commented,
    CommentedKV,
    CommentedList,
    CommentKey,
    IntermediateRepr,
    WhitespaceKey,
)

__all__ = [
    "dumps",
    "loads",
    "convert",
    "collapse",
]

T = TypeVar("T", bound=Union[TOMLDocument, Table, InlineTable])

MAX_INLINE_TABLE_LEN = 60
INLINE_TABLE_LONG_ELEM = 10
MAX_INLINE_TABLE_LONG_ELEM = 5
LONG = 120


def convert(irepr: IntermediateRepr) -> str:
    return dumps(collapse(irepr, root=True))


@singledispatch
def collapse(obj, root=False):
    # Catch all
    return obj


@collapse.register(Commented)
def _collapse_commented(obj: Commented, root=False) -> Item:
    return create_item(obj.value_or(None), obj.comment)


@collapse.register(CommentedList)
def _collapse_commented_list(obj: CommentedList, root=False) -> Array:
    out = array()
    out.multiline(False)  # Let's manually control the whitespace
    multiline = len(obj) > 1

    for entry in obj.data:
        values = entry.value_or([])
        if multiline:
            cast(list, out).append(Whitespace("\n" + 4 * " "))
        for value in values:
            cast(list, out).append(collapse(value))
        if entry.has_comment():
            if multiline:
                cast(list, out).append(_no_trail_comment(entry.comment))
            else:
                cast(Item, out).comment(entry.comment)
    if multiline:
        cast(list, out).append(Whitespace("\n"))

    return out


@collapse.register(CommentedKV)
def _collapse_commented_kv(obj: CommentedKV, root=False) -> Union[Table, InlineTable]:
    multiline = len(obj) > 1
    out: Union[Table, InlineTable] = table() if multiline else inline_table()

    for entry in obj.data:
        values = (v for v in entry.value_or([cast(KV, ())]) if v)
        k: Optional[str] = None  # if the for loop is empty, k = None
        for value in values:
            k, v = value
            out[cast(str, k)] = collapse(v)
        if not entry.has_comment():
            continue
        if not multiline:
            cast(Item, out).comment(entry.comment)
            obj.comment = entry.comment
            return out
        if k:
            out[k].comment(entry.comment)
        else:
            out.append(None, comment(entry.comment))  # type: ignore[arg-type]
    return out


@collapse.register(IntermediateRepr)
def _collapse_irepr(obj: IntermediateRepr, root=False):
    # guess a good repr
    if root:
        return _convert_irepr_to_toml(obj, document())
    rough_repr = repr(obj).replace(obj.__class__.__name__, "").strip()
    out = table() if len(rough_repr) > LONG or "\n" in rough_repr else inline_table()
    return _convert_irepr_to_toml(obj, cast(Union[Table, InlineTable], out))


@collapse.register(dict)
def _collapse_dict(obj: dict, root=False) -> Union[Table, InlineTable]:
    if not obj:
        return inline_table()
    out: Union[Table, InlineTable] = (
        table()
        if any(v and isinstance(v, (list, dict)) for v in obj.values())
        or len(repr(obj)) > LONG  # simple heuristic
        else inline_table()
    )
    for key, value in obj.items():
        out.append(key, collapse(value))
    return out


@collapse.register(list)
def _collapse_list(obj: list, root=False) -> Union[AoT, Array]:
    is_aot, max_len, has_nl, num_elem = classify_list(obj)
    # Just some heuristics which kind of array we are going to use
    if is_aot:
        return create_aot(collapse(e) for e in obj)

    out = array()
    cast(MutableSequence, out).extend(collapse(e) for e in obj)
    if (
        has_nl
        or max_len > MAX_INLINE_TABLE_LEN
        or (max_len > INLINE_TABLE_LONG_ELEM and num_elem > MAX_INLINE_TABLE_LONG_ELEM)
    ):
        out.multiline(True)
    return out


def _convert_irepr_to_toml(irepr: IntermediateRepr, out: T) -> T:
    if irepr.inline_comment and isinstance(out, Item):
        out.comment(irepr.inline_comment)
    for key, value in irepr.items():
        # TODO: prefer `add` once atoml's InlineTable supports it
        if isinstance(key, WhitespaceKey):
            out.append(None, nl())  # type: ignore[arg-type]
        elif isinstance(key, CommentKey):
            out.append(None, comment(value))  # type: ignore[arg-type]
        elif isinstance(key, tuple):
            parent_key, *rest = key
            if not isinstance(parent_key, str):
                raise InvalidTOMLKey(key)
            if len(rest) == 1:
                nested_key = rest[0]
                collapsed_value = collapse(value)
                collapsed_str = f"{nested_key} = {dumps(collapsed_value)}"
                # Force inline table for the simplest cases
                if (
                    not isinstance(collapsed_value, (Table, AoT))
                    and len(collapsed_str) < LONG
                    and "\n" not in collapsed_str.strip()
                ):
                    child = inline_table()
                    child[nested_key] = collapsed_value
                    out[parent_key] = child
                    continue
            else:
                nested_key = tuple(rest)
            p = out.setdefault(parent_key, {})
            if not isinstance(p, (Table, InlineTable)):
                msg = f"Value for `{parent_key}` expected to be a table, found {p!r}"
                raise ValueError(msg)
            _convert_irepr_to_toml(IntermediateRepr({nested_key: value}), p)
        elif isinstance(key, (int, str)):
            if isinstance(value, IntermediateRepr):
                _convert_irepr_to_toml(value, out.setdefault(str(key), {}))
            else:
                out[str(key)] = collapse(value)
    return out


# --- Helpers ---


def create_item(value, comment):
    obj = item(value)
    if comment is not None:
        obj.comment(comment)
    return obj


def create_table(m: Mapping) -> Table:
    if isinstance(m, Table):
        return m
    t = table()
    for k, v in m.items():
        t.add(k, v)
    return t


def create_aot(elements: Iterable[Mapping]) -> AoT:
    if isinstance(elements, AoT):
        return elements
    out = aot()
    cast(MutableSequence, out).extend(create_table(t) for t in elements)
    return out


def classify_list(seq: Sequence) -> Tuple[bool, int, bool, int]:
    """Expensive method that helps to choose what is the best TOML representation
    for a Python list.

    The return value is composed by 4 values, in order:
    - aot(bool): ``True`` if all elements are dict-like objects
    - max_len(int): Rough (and definitely not precise) estimative of the number of
      chars the TOML representation for the largest element would be.
    - has_nl(bool): if any TOML representation for the elements has a ``\\n`` char.
    - num_elements(int): number of elements in the list.
    """
    is_aot = True
    has_nl = False
    max_len = 0
    for elem in seq:
        is_aot = is_aot and isinstance(elem, Mapping)
        elem_repr = repr(elem)
        max_len = max(max_len, len(elem_repr))
        has_nl = has_nl or "\n" in elem_repr

    return is_aot, max_len, has_nl, len(seq)


def _no_trail_comment(msg: str):
    cmt = comment(str(msg))
    cmt.trivia.trail = ""
    return cmt



from ini2toml.drivers import configupdater as lib
from ini2toml.types import CommentKey
from ini2toml.types import IntermediateRepr as IR
from ini2toml.types import WhitespaceKey

example_cfg = """\
[section1]
# comment
value = 42 # int value

[section2]
float-value = 1.5
boolean-value = false
other value =
    1, 2, 3, # 1st line comment
    # 2nd line comment
other-boolean-value = true

# comment between options
string-value = string # comment

[section2.another value]
a = 1
b = 2 # 1st line comment
c = 3
d = 4 # 2nd line comment
"""


example_parsed = IR(
    {
        "section1": IR(
            {CommentKey(): "comment", "value": "42 # int value", WhitespaceKey(): "\n"}
        ),
        "section2": IR(
            {
                "float-value": "1.5",
                "boolean-value": "false",
                "other value": "\n1, 2, 3, # 1st line comment\n# 2nd line comment",
                "other-boolean-value": "true",
                WhitespaceKey(): "\n",
                CommentKey(): "comment between options",
                "string-value": "string # comment",
                WhitespaceKey(): "\n",
            }
        ),
        "section2.another value": IR(
            a="1", b="2 # 1st line comment", c="3", d="4 # 2nd line comment"
        ),
    }
)


def test_parse():
    assert lib.parse(example_cfg.strip(), {}) == example_parsed



from ini2toml.drivers import lite_toml as lib
from ini2toml.types import Commented, CommentedKV, CommentedList, CommentKey
from ini2toml.types import IntermediateRepr as IR
from ini2toml.types import WhitespaceKey

example_toml = """\
[section1]
value = 42

[section2]
float-value = 1.5
boolean-value = false
"other value" = [
    1,
    2,
    3,
]
other-boolean-value = true
string-value = "string"
list-value = [
    true,
    false,
]

[section2."another value"]
a = 1
b = 2
c = 3
d = 4

[section3.nested]
x = "y"
z = "w"
"""


example_parsed = IR(
    section1=IR({CommentKey(): "comment", "value": Commented(42, "int value")}),
    section2=IR(
        {
            "float-value": 1.5,
            "boolean-value": False,
            "other value": CommentedList(
                [
                    Commented([1, 2, 3], "1st line comment"),
                    Commented(comment="2nd line comment"),
                ]
            ),
            "other-boolean-value": True,
            WhitespaceKey(): "",
            CommentKey(): "comment between options",
            "another value": CommentedKV(
                [
                    Commented([("a", 1), ("b", 2)], "1st line comment"),
                    Commented([("c", 3), ("d", 4)], "2nd line comment"),
                ]
            ),
            "string-value": Commented("string", "comment"),
            "list-value": [True, False],
        }
    ),
    section3=IR(nested=IR(x="y", z=Commented("w", "nested"))),
)


def test_convert():
    assert lib.convert(example_parsed) == example_toml



# https://docs.pytest.org/en/latest/reference/reference.html#configuration-options
# https://docs.pytest.org/en/latest/reference/customize.html#config-file-formats
from collections.abc import MutableMapping
from functools import partial
from typing import TypeVar

from ..transformations import coerce_scalar, split_list
from ..types import IntermediateRepr, Translator

R = TypeVar("R", bound=IntermediateRepr)

list_with_space = partial(split_list, sep=" ")
split_markers = partial(split_list, sep="\n")
# ^ most of the list values in pytest use whitespace separators,
#   but markers are a special case, since they can define a help text


def activate(translator: Translator):
    plugin = Pytest()
    for file in ("setup.cfg", "tox.ini", "pytest.ini"):
        translator[file].intermediate_processors.append(plugin.process_values)
    translator["pytest.ini"].help_text = plugin.__doc__ or ""


class Pytest:
    """Convert settings to 'pyproject.toml' ('ini_options' table)"""

    LIST_VALUES = (
        "filterwarnings",
        "norecursedirs",
        "python_classes",
        "python_files",
        "python_functions",
        "required_plugins",
        "testpaths",
        "usefixtures",
    )

    DONT_TOUCH = ("minversion",)

    def process_values(self, doc: R) -> R:
        candidates = [
            (("pytest", "ini_options"), "pytest", doc),
            (("tool", "pytest", "ini_options"), "tool:pytest", doc),
            (("tool", "pytest", "ini_options"), ("tool", "pytest"), doc),
            (("pytest", "ini_options"), "pytest", doc.get("tool", {})),
        ]
        for new_key, old_key, parent in candidates:
            section = parent.get(old_key)
            if section:
                self.process_section(section)
                parent.rename(old_key, new_key)
        return doc

    def process_section(self, section: MutableMapping):
        for field in section:
            if field in self.DONT_TOUCH:
                continue
            if field == "markers":
                section[field] = split_markers(section[field])
            elif field in self.LIST_VALUES:
                section[field] = list_with_space(section[field])
            else:
                section[field] = coerce_scalar(section[field])



from configparser import ConfigParser
from types import MappingProxyType
from typing import Mapping

from ..types import IntermediateRepr

EMPTY: Mapping = MappingProxyType({})


def parse(text: str, opts: Mapping = EMPTY) -> IntermediateRepr:
    cfg = ConfigParser(**opts)
    cfg.read_string(text)
    irepr = IntermediateRepr()
    for name, section in cfg.items():
        if name == "DEFAULT":
            continue
        irepr.append(name, translate_section(section))
    return irepr


def translate_section(section: Mapping):
    irepr = IntermediateRepr()
    for name, value in section.items():
        irepr.append(name, value)
    return irepr



from ini2toml.drivers import plain_builtins as lib
from ini2toml.types import Commented, CommentedKV, CommentedList, CommentKey
from ini2toml.types import IntermediateRepr as IR
from ini2toml.types import WhitespaceKey

example_output = {
    "section1": {"value": 42},
    "section2": {
        "float-value": 1.5,
        "boolean-value": False,
        "other value": [1, 2, 3],
        "other-boolean-value": True,
        "string-value": "string",
        "list-value": [True, False],
        "another value": {"a": 1, "b": 2, "c": 3, "d": 4},
    },
    "section3": {"nested": {"x": "y", "z": "w"}},
}


example_parsed = IR(
    section1=IR({CommentKey(): "comment", "value": Commented(42, "int value")}),
    section2=IR(
        {
            "float-value": 1.5,
            "boolean-value": False,
            "other value": CommentedList(
                [
                    Commented([1, 2, 3], "1st line comment"),
                    Commented(comment="2nd line comment"),
                ]
            ),
            "other-boolean-value": True,
            WhitespaceKey(): "",
            CommentKey(): "comment between options",
            "another value": CommentedKV(
                [
                    Commented([("a", 1), ("b", 2)], "1st line comment"),
                    Commented([("c", 3), ("d", 4)], "2nd line comment"),
                ]
            ),
            "string-value": Commented("string", "comment"),
            "list-value": [True, False],
        }
    ),
    section3=IR(nested=IR(x="y", z=Commented("w", "nested"))),
)


def test_convert():
    assert lib.convert(example_parsed) == example_output



import re
from itertools import chain
from pathlib import Path

import pytest
import tomli
from validate_pyproject.api import Validator

from ini2toml import cli
from ini2toml.drivers import configparser, full_toml, lite_toml
from ini2toml.translator import Translator


def examples():
    here = Path(".").resolve()
    parent = Path(__file__).parent / "examples"
    for folder in parent.glob("*/"):
        cfg = chain(folder.glob("*.cfg"), folder.glob("*.ini"))
        toml = folder.glob("*.toml")
        for orig in cfg:
            expected = orig.with_suffix(".toml")
            if expected.is_file():
                yield str(orig.relative_to(here)), str(expected.relative_to(here))
            else:
                try:
                    yield str(orig.relative_to(here)), str(next(toml).relative_to(here))
                except:  # noqa
                    print(f"Missing TOML file to compare to {orig}")
                    raise


@pytest.fixture(scope="module")
def validate():
    """Use ``validate-pyproject`` to validate the generated TOML"""
    return Validator()


@pytest.mark.parametrize(("original", "expected"), list(examples()))
def test_examples_api(original, expected, validate):
    translator = Translator()
    available_profiles = list(translator.profiles.keys())
    profile = cli.guess_profile(None, original, available_profiles)
    out = translator.translate(Path(original).read_text(), profile)
    expected_text = Path(expected).read_text().strip()
    assert out == expected_text
    # Make sure they can be parsed
    dict_equivalent = tomli.loads(out)
    assert dict_equivalent == tomli.loads(expected_text)
    assert validate(remove_deprecated(dict_equivalent)) is not None


COMMENT_LINE = re.compile(r"^\s*#[^\n]*\n", re.M)
INLINE_COMMENT = re.compile(r'#\s[^\n"]*$', re.M)


@pytest.mark.parametrize(("original", "expected"), list(examples()))
def test_examples_api_lite(original, expected, validate):
    opts = {"ini_loads_fn": configparser.parse, "toml_dumps_fn": lite_toml.convert}
    translator = Translator(**opts)
    available_profiles = list(translator.profiles.keys())
    profile = cli.guess_profile(None, original, available_profiles)
    # We cannot compare "flake8" sections (currently not handled)
    # (ConfigParser automatically strips comments, contrary to ConfigUpdater)
    orig = Path(original)
    out = remove_flake8_from_toml(translator.translate(orig.read_text(), profile))
    expected_text = remove_flake8_from_toml(Path(expected).read_text().strip())

    # At least the Python-equivalents should be the same when parsing
    dict_equivalent = tomli.loads(out)
    assert dict_equivalent == tomli.loads(expected_text)
    assert validate(remove_deprecated(dict_equivalent)) is not None

    without_comments = COMMENT_LINE.sub("", expected_text)
    without_comments = INLINE_COMMENT.sub("", without_comments)
    try:
        assert out == without_comments
    except AssertionError:
        # We can ignore some minor formatting differences, as long as the parsed
        # dict is the same
        pass


@pytest.mark.parametrize(("original", "expected"), list(examples()))
def test_examples_cli(original, expected, capsys):
    cli.run([original])
    (out, err) = capsys.readouterr()
    expected_text = Path(expected).read_text().strip()
    assert out == expected_text
    # Make sure they can be parsed
    assert tomli.loads(out) == tomli.loads(expected_text)


def remove_flake8_from_toml(text: str) -> str:
    # full_toml uses atoml that should not change any formatting, just remove the
    # parts we don't want
    doc = full_toml.loads(text)
    tool = doc.get("tool", {})
    for key in list(tool.keys()):  # eager to allow dictionary modifications
        if key.startswith("flake8"):
            tool.pop(key)
    return full_toml.dumps(doc)


def remove_deprecated(dict_equivalent):
    setuptools = dict_equivalent.get("tool", {}).get("setuptools", {})
    # The usage of ``data-files`` is deprecated in setuptools and not supported by pip
    setuptools.pop("data-files", None)
    return dict_equivalent



import logging
from typing import Mapping, Optional, Sequence

from . import types  # Structural/Abstract types
from .base_translator import EMPTY, BaseTranslator
from .plugins import list_from_entry_points as list_all_plugins

_logger = logging.getLogger(__name__)


class Translator(BaseTranslator[str]):
    """``Translator`` is the main public Python API exposed by the ``ini2toml``,
    to convert strings represeting ``.ini/.cfg`` files into the ``TOML`` syntax.

    It follows the public API defined in :class:`ini2toml.types.Translator`, and is very
    similar to :class:`~ini2toml.base_translator.BaseTranslator`.
    The main difference is that ``Translator`` will try to automatically figure out the
    values for ``plugins``, ``ini_loads_fn`` and ``toml_dumps_fn`` when they are not
    passed, based on the installed Python packages (and available `entry-points`_),
    while ``BaseTranslator`` requires the user to explicitly set these parameters.

    For most of the users ``Translator`` is recommended over ``BaseTranslator``.

    See :class:`~ini2toml.base_translator.BaseTranslator` for a description of the
    instantiation parameters.

    .. _entry-points: https://packaging.python.org/specifications/entry-points/
    """

    def __init__(
        self,
        plugins: Optional[Sequence[types.Plugin]] = None,
        profiles: Sequence[types.Profile] = (),
        profile_augmentations: Sequence[types.ProfileAugmentation] = (),
        ini_parser_opts: Mapping = EMPTY,
        ini_loads_fn: Optional[types.IniLoadsFn] = None,
        toml_dumps_fn: Optional[types.TomlDumpsFn] = None,
    ):
        super().__init__(
            ini_loads_fn=ini_loads_fn or _discover_ini_loads_fn(),
            toml_dumps_fn=toml_dumps_fn or _discover_toml_dumps_fn(),
            plugins=list_all_plugins() if plugins is None else plugins,
            ini_parser_opts=ini_parser_opts,
            profiles=profiles,
            profile_augmentations=profile_augmentations,
        )


def _discover_ini_loads_fn() -> types.IniLoadsFn:
    try:
        from .drivers.configupdater import parse
    except ImportError:  # pragma: no cover
        from .drivers.configparser import parse  # type: ignore[no-redef]

    return parse


def _discover_toml_dumps_fn() -> types.TomlDumpsFn:
    try:
        from .drivers.full_toml import convert
    except ImportError:  # pragma: no cover
        try:
            from .drivers.lite_toml import convert  # type: ignore[no-redef]
        except ImportError:
            msg = "Please install either `ini2toml[full]` or `ini2toml[lite]`"
            _logger.warning(f"{msg}. `ini2toml` (alone) is not valid.")
            raise

    return convert



"""Intermediate representations used by ``ini2toml`` when transforming between
the INI and TOML syntaxes.
"""
from collections import UserList
from dataclasses import dataclass, field
from enum import Enum
from itertools import chain
from pprint import pformat
from textwrap import indent
from types import MappingProxyType
from typing import (
    Any,
    Dict,
    Generic,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
)
from uuid import uuid4

T = TypeVar("T")
S = TypeVar("S")
R = TypeVar("R", bound="IntermediateRepr")

KV = Tuple[str, T]

NotGiven = Enum("NotGiven", "NOT_GIVEN")
NOT_GIVEN = NotGiven.NOT_GIVEN

EMPTY: Mapping = MappingProxyType({})


@dataclass(frozen=True)
class HiddenKey:
    _value: int = field(default_factory=lambda: uuid4().int)

    def __str__(self):
        return f"{self.__class__.__name__}()"

    __repr__ = __str__


class WhitespaceKey(HiddenKey):
    pass


class CommentKey(HiddenKey):
    pass


Key = Union[str, HiddenKey, Tuple[Union[str, HiddenKey], ...]]


class IntermediateRepr(MutableMapping):
    def __init__(
        self,
        elements: Mapping[Key, Any] = EMPTY,
        order: Sequence[Key] = (),
        inline_comment: str = "",
        **kwargs,
    ):
        el = chain(elements.items(), kwargs.items())
        self.elements: Dict[Key, Any] = {}
        self.order: List[Key] = []
        self.inline_comment = inline_comment
        self.elements.update(el)
        self.order.extend(order or self.elements.keys())
        elem_not_in_order = any(k not in self.order for k in self.elements)
        order_not_in_elem = any(k not in self.elements for k in self.order)
        if elem_not_in_order or order_not_in_elem:
            raise ValueError(f"{order} and {elements} need to have the same keys")

    def __repr__(self):
        inner = ",\n".join(
            indent(f"{k}={pformat(getattr(self, k))}", "    ")
            for k in ("elements", "order", "inline_comment")
        )
        return f"{self.__class__.__name__}(\n{inner}\n)"

    def __eq__(self, other):
        L = len(self)
        if not (
            isinstance(other, self.__class__)
            and self.inline_comment == other.inline_comment
            and len(other) == L
        ):
            return False
        self_ = [(str(k), v) for k, v in self.items()]
        other_ = [(str(k), v) for k, v in other.items()]
        return all(self_[i] == other_[i] for i in range(L))

    def rename(self, old_key: Key, new_key: Key, ignore_missing=False):
        """This method renames an existing key, without changing its position.
        Notice that ``new_key`` cannot be already present, and that trying to rename
        a non-pre-existing key will also result in error (unless
        ``ignore_missing=True``).
        """
        if old_key == new_key:
            return self
        if new_key in self.order:
            raise KeyError(f"new_key={new_key!r} already exists")
        if old_key not in self.order and ignore_missing:
            return self
        i = self.order.index(old_key)
        self.order[i] = new_key
        self.elements[new_key] = self.elements.pop(old_key)
        return self

    def insert(self, position: int, key: Key, value: Any):
        """Simulate the position-aware :meth:`collections.abc.MutableMapping.insert`
        method, but also require a ``key`` to be specified.
        """
        if key in self.order:
            raise KeyError(f"key={key!r} already exists")
        self.order.insert(position, key)
        self.elements[key] = value

    def index(self, key: Key) -> int:
        """Find the position of ``key``"""
        return self.order.index(key)

    def append(self, key: Key, value: Any):
        """Simulate the position-aware :meth:`collections.abc.MutableMapping.append`
        method, but also require a ``key`` to be specified.
        """
        self.insert(len(self.order), key, value)

    def copy(self: R) -> R:
        return self.__class__(self.elements.copy(), self.order[:], self.inline_comment)

    def replace_first_remove_others(
        self, existing_keys: Sequence[Key], new_key: Key, value: Any
    ):
        """Find the first key in ``existing_keys`` that existing in the intermediate
        representation, and replaces it with ``new_key`` (similar to
        :meth:`replace`).
        All the other keys in ``existing_keys`` are removed and the value of
        ``new_key`` is set to ``value``.
        """
        idx = [self.index(k) for k in existing_keys if k in self]
        if not idx:
            i = len(self)
        else:
            i = sorted(idx)[0]
            for key in existing_keys:
                self.pop(key, None)
        self.insert(i, new_key, value)
        return i

    def __getitem__(self, key: Key):
        return self.elements[key]

    def __setitem__(self, key: Key, value: Any):
        if key not in self.elements:
            self.order.append(key)
        self.elements[key] = value

    def __delitem__(self, key: Key):
        del self.elements[key]
        self.order.remove(key)

    def __iter__(self):
        return iter(self.order)

    def __len__(self):
        return len(self.order)


# These objects hold information about the processed values + comments
# in such a way that we can later convert them to TOML while still preserving
# the comments (if we want to).


@dataclass
class Commented(Generic[T]):
    value: Union[T, NotGiven] = field(default_factory=lambda: NOT_GIVEN)
    comment: Optional[str] = field(default_factory=lambda: None)

    def comment_only(self):
        return self.value is NOT_GIVEN

    def has_comment(self):
        return bool(self.comment)

    def value_or(self, fallback: S) -> Union[T, S]:
        return fallback if self.value is NOT_GIVEN else self.value


class CommentedList(Generic[T], UserList):
    def __init__(self, data: List[Commented[List[T]]]):
        super().__init__(data)
        self.comment: Optional[str] = None  # TODO: remove this workaround

    def as_list(self) -> list:
        out = []
        for entry in self:
            values = entry.value_or([])
            for value in values:
                out.append(value)
        return out


class CommentedKV(Generic[T], UserList):
    def __init__(self, data: List[Commented[List[KV[T]]]]):
        super().__init__(data)
        self.comment: Optional[str] = None  # TODO: remove this workaround

    def find(self, key: str) -> Optional[Tuple[int, int]]:
        for i, row in enumerate(self):
            for j, item in enumerate(row.value_or([])):
                if item[0] == key:
                    return (i, j)
        return None

    def as_dict(self) -> dict:
        out = {}
        for entry in self:
            values = (v for v in entry.value_or([cast(KV, ())]) if v)
            for k, v in values:
                out[k] = v
        return out

    def to_ir(self) -> IntermediateRepr:
        """:class:`CommentedKV` are usually intended to represent INI options, while
        :class:`IntermediateRepr` are usually intended to represent INI sections.
        Therefore this function allows "promoting" an option-equivalent to a
        section-equivalent representation.
        """
        irepr = IntermediateRepr()
        for row in self:
            for key, value in row.value_or([]):
                irepr[key] = value
            if row.has_comment():
                irepr[key] = Commented(value, row.comment)

        return irepr



import logging
import sys
from unittest.mock import MagicMock

import pytest

from ini2toml import cli
from ini2toml.profile import Profile, ProfileAugmentation


def test_guess_profile(caplog):
    available = ["best_effort", "setup.cfg", "mypy.ini", "pyscaffold/default.cfg"]
    with caplog.at_level(logging.DEBUG):
        assert cli.guess_profile("best_effort", "setup.cfg", available) == "best_effort"
        assert caplog.text == ""
    caplog.clear()
    with caplog.at_level(logging.DEBUG):
        assert cli.guess_profile(None, "myproj/setup.cfg", available) == "setup.cfg"
        assert "Profile not explicitly set" in caplog.text
        assert "'setup.cfg' inferred" in caplog.text
    caplog.clear()
    with caplog.at_level(logging.DEBUG):
        profile = cli.guess_profile(None, "~/.config/pyscaffold/default.cfg", available)
        assert profile == "pyscaffold/default.cfg"
        assert "Profile not explicitly set" in caplog.text
        assert "'pyscaffold/default.cfg' inferred" in caplog.text
    caplog.clear()
    with caplog.at_level(logging.DEBUG):
        profile = cli.guess_profile(None, "myfile.ini", available)
        assert "Profile not explicitly set" in caplog.text
        assert "using 'best_effort'" in caplog.text


def _profile(name, help_text="help_text"):
    return Profile(name, help_text)


def _aug(name, active, help_text="help_text"):
    return ProfileAugmentation(lambda x: x, active, name, help_text)


def test_parse_args(tmp_path):
    file = tmp_path / "file.ini"
    file.touch()
    args = f"{file} -p setup.cfg -D hello world -E ehllo owrld".split()
    profiles = [_profile(n) for n in "setup.cfg default.cfg".split()]
    aug = [_aug(n, True) for n in "hello world other".split()]
    aug += [_aug(n, False) for n in "ehllo owrld".split()]
    params = cli.parse_args(args, profiles, aug)
    assert params.profile == "setup.cfg"
    assert params.active_augmentations == {
        "hello": False,
        "world": False,
        "ehllo": True,
        "owrld": True,
    }
    assert "other" not in params.active_augmentations


def test_critical_logging_sets_log_level_on_error(monkeypatch, caplog):
    spy = MagicMock()
    monkeypatch.setattr(sys, "argv", ["-vv"])
    monkeypatch.setattr(logging, "basicConfig", spy)
    with pytest.raises(ValueError):
        with cli.critical_logging():
            raise ValueError
    _args, kwargs = spy.call_args
    assert kwargs["level"] == logging.DEBUG


def test_critical_logging_does_nothing_if_no_argv(monkeypatch, caplog):
    spy = MagicMock()
    monkeypatch.setattr(sys, "argv", [])
    monkeypatch.setattr(logging, "basicConfig", spy)
    with pytest.raises(ValueError):
        with cli.critical_logging():
            raise ValueError
    assert spy.call_args is None


def test_exceptisons2exit():
    with pytest.raises(SystemExit):
        with cli.exceptisons2exit():
            raise ValueError


def test_help(capsys):
    with pytest.raises(SystemExit):
        cli.run(["--help"])
    out, _ = capsys.readouterr()
    assert len(out) > 0
    text = " ".join(x.strip() for x in out.splitlines())
    expected_profile_desc = """\
    conversion. Available profiles:
    - "best_effort": guess option value conversion based
    on the string format.
    - ".coveragerc": convert settings to 'pyproject.toml'
    equivalent.
    - "setup.cfg": convert settings to 'pyproject.toml'
    based on PEP 621.
    - ".isort.cfg": convert settings to 'pyproject.toml'
    equivalent.
    - "mypy.ini": convert settings to 'pyproject.toml'
    equivalent.
    - "pytest.ini": convert settings to 'pyproject.toml'
    ('ini_options' table).
    """
    expected = " ".join(x.strip() for x in expected_profile_desc.splitlines())
    assert expected in text



"""Reusable value and type casting transformations"""
import logging
from collections.abc import MutableMapping
from functools import reduce, wraps
from typing import (
    Any,
    Callable,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)

from .types import Commented, CommentedKV, CommentedList

CP = ("#", ";")
"""Default Comment Prefixes"""

S = TypeVar("S")
T = TypeVar("T")
U = TypeVar("U")
V = TypeVar("V")
X = TypeVar("X")
Y = TypeVar("Y")
M = TypeVar("M", bound=MutableMapping)
KV = Tuple[str, T]

FN = Callable[[X], Y]

Scalar = Union[int, float, bool, str]  # TODO: missing time and datetime
"""Simple data types with TOML correspondence"""

CoerceFn = Callable[[str], T]
"""Functions that know how to parser/coerce string values into different types"""

Transformation = Union[Callable[[str], Any], Callable[[M], M]]
"""There are 2 main types of transformation:

- The first one is a simple transformation that processes a string value (coming from an
  option in the original CFG/INI file) into a value with an equivalent TOML data type.
  For example: transforming ``"2"`` (string) into ``2`` (integer).
- The second one tries to preserve metadata (such as comments) from the original CFG/INI
  file. This kind of transformation processes a string value into an intermediary
  representation (e.g. :obj:`Commented`, :obj:`CommentedList`, obj:`CommentedKV`)
  that needs to be properly handled before adding to the TOML document.

In a higher level we can also consider an ensemble of transformations that transform an
entire table of the TOML document.
"""

TF = TypeVar("TF", bound=Transformation)

_logger = logging.getLogger(__name__)


# ---- Simple value processors ----
# These functions return plain objects, that can be directly added to the TOML document


def noop(x: T) -> T:
    return x


def is_true(value: str) -> bool:
    value = value.lower()
    return value in ("true", "1", "yes", "on")


def is_false(value: str) -> bool:
    value = value.lower()
    return value in ("false", "0", "no", "off", "none", "null", "nil")


def is_float(value: str) -> bool:
    cleaned = value.lower().lstrip("+-").replace(".", "").replace("_", "")
    return cleaned.isdecimal() and value.count(".") <= 1 or cleaned in ("inf", "nan")


def coerce_bool(value: str) -> bool:
    if is_true(value):
        return True
    if is_false(value):
        return False
    raise ValueError(f"{value!r} cannot be converted to boolean")


def coerce_scalar(value: str) -> Scalar:
    """Try to convert the given string to a proper "scalar" type (e.g. integer, float,
    bool, ...) with an direct TOML equivalent.
    If the conversion is unknown or not possible, it will return the same input value
    (as string).

    .. note:: This function "guesses" the value type based in heuristics and/or regular
       expressions, therefore there is no guarantee the output has the same type as
       intended by the original author.

    .. note:: Currently date/time-related types are not supported.
    """
    value = value.strip()
    if value.isdecimal():
        return int(value)
    if is_float(value):
        return float(value)
    if is_true(value):
        return True
    elif is_false(value):
        return False
    # Do we need this? Or is there a better way? How about time objects
    # > try:
    # >     return datetime.fromisoformat(value)
    # > except ValueError:
    # >     pass
    return value


def kebab_case(field: str) -> str:
    return field.lower().replace("_", "-")


def deprecated(
    name: str, fn: TF = noop, instead: str = ""  # type: ignore[assignment]
) -> TF:
    """Wrapper around the ``fn`` transformation to warn user about deprecation."""
    extra = f". Use {instead!r} instead" if instead else ""

    @wraps(fn)
    def _fn(*args, **kwargs):
        _logger.warning(f"{name!r} is deprecated{extra}.")
        return fn(*args, **kwargs)

    return cast(TF, _fn)


# ---- Complex value processors ----
# These functions return an intermediate representation of the value,
# that need `apply` to be added to a container


@overload
def split_comment(value: str, *, comment_prefixes=CP) -> Commented[str]:
    ...


@overload
def split_comment(
    value: str, coerce_fn: CoerceFn[T], comment_prefixes=CP
) -> Commented[T]:
    ...


def split_comment(value, coerce_fn=noop, comment_prefixes=CP):
    if not isinstance(value, str):
        return value
    value = value.strip()
    prefixes = [p for p in comment_prefixes if p in value]

    # We just process inline comments for single line options
    if not prefixes or len(value.splitlines()) > 1:
        return Commented(coerce_fn(value))

    if any(value.startswith(p) for p in comment_prefixes):
        return Commented(comment=_strip_comment(value, comment_prefixes))

    prefix = prefixes[0]  # We can only analyse one...
    value, cmt = _split_in_2(value, prefix)
    return Commented(coerce_fn(value.strip()), _strip_comment(cmt, comment_prefixes))


def split_scalar(value: str, *, comment_prefixes=CP) -> Commented[Scalar]:
    return split_comment(value, coerce_scalar, comment_prefixes)


@overload
def split_list(
    value: str, sep: str = ",", *, subsplit_dangling=True, comment_prefixes=CP
) -> CommentedList[str]:
    ...


@overload
def split_list(
    value: str,
    sep: str = ",",
    *,
    coerce_fn: CoerceFn[T],
    subsplit_dangling=True,
    comment_prefixes=CP,
) -> CommentedList[T]:
    ...


@overload
def split_list(
    value: str,
    sep: str,
    coerce_fn: CoerceFn[T],
    subsplit_dangling=True,
    comment_prefixes=CP,
) -> CommentedList[T]:
    ...


def split_list(
    value, sep=",", coerce_fn=noop, subsplit_dangling=True, comment_prefixes=CP
):
    """Value encoded as a (potentially) dangling list values separated by ``sep``.

    This function will first try to split the value by lines (dangling list) using
    :func:`str.splitlines`. Then, if ``subsplit_dangling=True``, it will split each line
    using ``sep``. As a result a list of items is obtained.
    For each item in this list ``coerce_fn`` is applied.
    """
    if not isinstance(value, str):
        return value
    comment_prefixes = [p for p in comment_prefixes if sep not in p]

    values = value.strip().splitlines()
    if not subsplit_dangling and len(values) > 1:
        sep += "\n"  # force a pattern that cannot be found in a split line

    def _split(line: str) -> list:
        return [coerce_fn(v.strip()) for v in line.split(sep) if v]

    return CommentedList([split_comment(v, _split, comment_prefixes) for v in values])


@overload
def split_kv_pairs(
    value: str,
    key_sep: str = "=",
    *,
    pair_sep=",",
    subsplit_dangling=True,
    comment_prefixes=CP,
) -> CommentedKV[str]:
    ...


@overload
def split_kv_pairs(
    value: str,
    key_sep: str = "=",
    *,
    coerce_fn: CoerceFn[T],
    pair_sep=",",
    subsplit_dangling=True,
    comment_prefixes=CP,
) -> CommentedKV[T]:
    ...


@overload
def split_kv_pairs(
    value: str,
    key_sep: str,
    coerce_fn: CoerceFn[T],
    pair_sep=",",
    subsplit_dangling=True,
    comment_prefixes=CP,
) -> CommentedKV[T]:
    ...


def split_kv_pairs(
    value,
    key_sep="=",
    coerce_fn=noop,
    pair_sep=",",
    subsplit_dangling=True,
    comment_prefixes=CP,
):
    """Value encoded as a (potentially) dangling list of key-value pairs.

    This function will first try to split the value by lines (dangling list) using
    :func:`str.splitlines`. Then, if ``subsplit_dangling=True``, it will split each line
    using ``pair_sep``. As a result a list of key-value pairs is obtained.
    For each item in this list, the key is separated from the value by ``key_sep``.
    ``coerce_fn`` is used to convert the value in each pair.
    """
    prefixes = [p for p in comment_prefixes if key_sep not in p and pair_sep not in p]

    values = value.strip().splitlines()
    if not subsplit_dangling and len(values) > 1:
        pair_sep += "\n"  # force a pattern that cannot be found in a split line

    def _split_kv(line: str) -> List[KV]:
        pairs = (
            item.split(key_sep, maxsplit=1)
            for item in line.strip().split(pair_sep)
            if key_sep in item
        )
        return [(k.strip(), coerce_fn(v.strip())) for k, v in pairs]

    return CommentedKV([split_comment(v, _split_kv, prefixes) for v in values])


# ---- Public Helpers ----


def remove_prefixes(text: str, prefixes: Sequence[str]):
    text = text.strip()
    for prefix in prefixes:
        if prefix and text.startswith(prefix):
            return text[len(prefix) :].strip()
    return text


def apply(x, fn):
    """Useful to reduce over a list of functions"""
    return fn(x)


@overload
def pipe(fn1: FN[S, T], fn2: FN[T, U]) -> FN[S, U]:
    ...


@overload
def pipe(fn1: FN[S, T], fn2: FN[T, U], fn3: FN[U, V]) -> FN[S, V]:
    ...


@overload
def pipe(fn1: FN[S, T], fn2: FN[T, U], fn3: FN[U, V], fn4: FN[V, X]) -> FN[S, X]:
    ...


@overload
def pipe(
    fn1: FN[S, T], fn2: FN[T, U], fn3: FN[U, V], fn4: FN[V, X], fn5: FN[X, Y]
) -> FN[S, Y]:
    ...


@overload
def pipe(
    fn1: FN[S, T],
    fn2: FN[T, U],
    fn3: FN[U, V],
    fn4: FN[V, X],
    fn5: FN[X, Y],
    *fn: FN[Y, Y],
) -> FN[S, Y]:
    ...


def pipe(*fns):
    """Compose 1-argument functions respecting the sequence they should be applied:

    .. code-block:: python

        pipe(fn1, fn2, fn3, ..., fnN)(x) == fnN(...(fn3(fn2(fn1(x)))))
    """
    return lambda x: reduce(apply, fns, x)


# ---- Private Helpers ----


def _split_in_2(v: str, sep: str) -> Tuple[str, Optional[str]]:
    items = iter(v.split(sep, maxsplit=1))
    first = next(items)
    second = next(items, None)
    return first, second


def _strip_comment(msg: Optional[str], prefixes: Sequence[str] = CP) -> Optional[str]:
    if not msg:
        return None
    return remove_prefixes(msg, prefixes)



from functools import partial
from textwrap import dedent

import pytest

from ini2toml import transformations as lib
from ini2toml.drivers.configupdater import parse as loads
from ini2toml.drivers.full_toml import convert as dumps


def test_coerce_bool():
    for value in ("true", "1", "yes", "on"):
        assert lib.coerce_bool(value) is True
        assert lib.coerce_bool(value.upper()) is True

    for value in ("false", "0", "no", "off", "none", "null", "nil"):
        assert lib.coerce_bool(value) is False
        assert lib.coerce_bool(value.upper()) is False

    with pytest.raises(ValueError):
        lib.coerce_bool("3")


def test_split_comment():
    example = "1 # comment"
    assert lib.split_comment(example) == lib.Commented("1", "comment")
    assert lib.split_comment(example, int) == lib.Commented(1, "comment")
    item = lib.split_comment(" # comment only")
    assert item.comment_only()


def test_split_list():
    example = "1, 2, 3 # comment"
    assert lib.split_list(example) == [lib.Commented(["1", "2", "3"], "comment")]
    example = "1, 2, 3 # comment"
    expected = [lib.Commented([1, 2, 3], "comment")]
    assert lib.split_list(example, coerce_fn=int) == expected
    example = "    1, 2, 3 # comment\n 4, 5, 6 # comment\n"
    expected = [
        lib.Commented([1, 2, 3], "comment"),
        lib.Commented([4, 5, 6], "comment"),
    ]
    assert lib.split_list(example, coerce_fn=int) == expected


def test_split_kv_pairs():
    example = "a=1, b=2, c=3 # comment"
    expected = [lib.Commented([("a", "1"), ("b", "2"), ("c", "3")], "comment")]
    assert lib.split_kv_pairs(example) == expected

    expected = [lib.Commented([("a", 1), ("b", 2), ("c", 3)], "comment")]
    assert lib.split_kv_pairs(example, coerce_fn=int) == expected

    example = "    a=1, b = 2, c =3 # comment\n d=4, e=5, f=6 # comment\n"
    expected = [
        lib.Commented([("a", 1), ("b", 2), ("c", 3)], "comment"),
        lib.Commented([("d", 4), ("e", 5), ("f", 6)], "comment"),
    ]
    assert lib.split_kv_pairs(example, coerce_fn=int) == expected


# The following tests are more of "integration tests" since they also use drivers


def apply(container, field, fn):
    container[field] = fn(container[field])
    return container


def test_application():
    example = """\
    [table]
    option1 = 1
    option2 = value # comment

    # commented single line compound value
    option3 = 1, 2, 3 # comment
    option4 = a=1, b=2, c=3 # comment

    # commented multiline compound value
    option5 =
        a=1 # comment
        b=2, c=3 # comment
    option5.1 =
        # header comment
        b=2, c=3 # comment
    option6 =
        1
        2 # comment
        3
    option7 =
        # comment
        1
        2

    # No subsplit dangling
    option8 =
        1, 2
        3
    option9 =
        1, 2
        3
    option10 =
        a=1
        b=2, c=3
    """

    doc = loads(dedent(example))
    split_int = partial(lib.split_list, coerce_fn=int)
    split_kv_int = partial(lib.split_kv_pairs, coerce_fn=int)
    dangling_list_no_subsplit = partial(lib.split_list, subsplit_dangling=False)
    dangling_kv_no_subsplit = partial(lib.split_kv_pairs, subsplit_dangling=False)

    doc["table"] = apply(doc["table"], "option1", int)
    expected = "option1 = 1"
    assert expected in dumps(doc)

    # assert len(_trailing_nl()) == 1

    doc["table"] = apply(doc["table"], "option2", lib.split_comment)
    expected = 'option2 = "value" # comment'
    assert expected in dumps(doc)

    doc["table"] = apply(doc["table"], "option3", split_int)
    expected = "option3 = [1, 2, 3] # comment"
    assert expected in dumps(doc)

    doc["table"] = apply(doc["table"], "option4", split_kv_int)
    expected = "option4 = {a = 1, b = 2, c = 3} # comment"
    assert expected in dumps(doc)

    doc["table"] = apply(doc["table"], "option5", split_kv_int)
    expected = """\
    [table.option5]
    a = 1 # comment
    b = 2
    c = 3 # comment
    """
    assert dedent(expected) in dumps(doc)

    doc["table"] = apply(doc["table"], "option5.1", split_kv_int)
    expected = """\
    [table."option5.1"]
    # header comment
    b = 2
    c = 3 # comment
    """
    assert dedent(expected) in dumps(doc)

    doc["table"] = apply(doc["table"], "option6", split_int)
    expected = """\
    option6 = [
        1, 
        2, # comment
        3, 
    ]
    """
    assert dedent(expected) in dumps(doc)

    doc["table"] = apply(doc["table"], "option7", split_int)
    expected = """\
    option7 = [
        # comment
        1, 
        2, 
    ]
    """
    assert dedent(expected) in dumps(doc)

    doc["table"] = apply(doc["table"], "option8", dangling_list_no_subsplit)
    expected = """\
    option8 = [
        "1, 2", 
        "3", 
    ]
    """
    assert dedent(expected) in dumps(doc)

    doc["table"] = apply(doc["table"], "option9", split_int)
    expected = """\
    option9 = [
        1, 2, 
        3, 
    ]
    """
    assert dedent(expected) in dumps(doc)

    doc["table"] = apply(doc["table"], "option10", dangling_kv_no_subsplit)
    expected = """\
    [table.option10]
    a = "1"
    b = "2, c=3"
    """
    assert dedent(expected) in dumps(doc)



from textwrap import dedent

import tomli

from ini2toml.drivers import full_toml, lite_toml
from ini2toml.plugins import pytest
from ini2toml.translator import Translator


def test_pytest():
    example = """\
    [pytest]
    minversion = 6.0
    addopts = -ra -q --cov ini2toml
    testpaths = tests
    python_files = test_*.py check_*.py example_*.py
    required_plugins = pytest-django>=3.0.0,<4.0.0 pytest-html pytest-xdist>=1.0.0
    norecursedirs =
        dist
        build
        .tox
    """
    expected = """\
    [pytest]
    [pytest.ini_options]
    minversion = "6.0"
    addopts = "-ra -q --cov ini2toml"
    testpaths = ["tests"]
    python_files = ["test_*.py", "check_*.py", "example_*.py"]
    required_plugins = ["pytest-django>=3.0.0,<4.0.0", "pytest-html", "pytest-xdist>=1.0.0"]
    norecursedirs = [
        "dist", 
        "build", 
        ".tox", 
    ]
    """  # noqa
    for convert in (lite_toml.convert, full_toml.convert):
        translator = Translator(plugins=[pytest.activate], toml_dumps_fn=convert)
        out = translator.translate(dedent(example), "pytest.ini").strip()
        expected = dedent(expected).strip()
        print("expected=\n" + expected + "\n***")
        print("out=\n" + out)
        try:
            assert expected in out
        except AssertionError:
            # At least the Python-equivalents when parsing should be the same
            assert tomli.loads(expected) == tomli.loads(out)



from textwrap import dedent

import pytest

from ini2toml.errors import UndefinedProfile
from ini2toml.translator import Translator


def test_simple_example():
    example = """\
    # comment

    [section1]
    option1 = value
    option2 = value # option comments are considered part of the value

    # comment
    [section2] # inline comment
    # comment
    option3 = value
    [section3]
    """
    # Obs: TOML always add a space before a new section
    expected = """\
    # comment

    [section1]
    option1 = "value"
    option2 = "value # option comments are considered part of the value"

    # comment

    [section2] # inline comment
    # comment
    option3 = "value"

    [section3]
    """
    translator = Translator(plugins=[])
    # ensure profile exists
    translator["simple"]
    out = translator.translate(dedent(example), "simple")
    print(out)
    assert out == dedent(expected)


def test_parser_opts():
    example = """\
    : comment

    [section1]
    option1 - value
    option2 - value : option comments are considered part of the value

    : comment
    [section2] : inline comment
    : comment
    option3 - value
    [section3]
    """
    # Obs: TOML always add a space before a new section
    expected = """\
    # comment

    [section1]
    option1 = "value"
    option2 = "value : option comments are considered part of the value"

    # comment

    [section2] # inline comment
    # comment
    option3 = "value"

    [section3]
    """

    parser_opts = {"comment_prefixes": (":",), "delimiters": ("-",)}
    translator = Translator(plugins=[], ini_parser_opts=parser_opts)
    # ensure profile exists
    translator["simple"]
    out = translator.translate(dedent(example), "simple")
    print(out)
    assert out == dedent(expected)


def test_undefined_profile():
    translator = Translator()
    with pytest.raises(UndefinedProfile):
        translator.translate("", "!!--UNDEFINED ??? PROFILE--!!")



import argparse
import logging
import sys
from contextlib import contextmanager
from itertools import chain
from os import pathsep
from pathlib import Path
from textwrap import dedent, wrap
from typing import Dict, List, Optional, Sequence

from . import __version__
from .translator import Translator
from .types import CLIChoice, Profile, ProfileAugmentation

_logger = logging.getLogger(__package__)

DEFAULT_PROFILE = "best_effort"


@contextmanager
def critical_logging():
    """Make sure the logging level is set even before parsing the CLI args"""
    try:
        yield
    except Exception:
        if "-vv" in sys.argv or "--very-verbose" in sys.argv:
            setup_logging(logging.DEBUG)
        raise


META: Dict[str, dict] = {
    "version": dict(
        flags=("-V", "--version"),
        action="version",
        version=f"{__package__} {__version__}",
    ),
    "input_file": dict(
        dest="input_file",
        type=argparse.FileType("r"),
        help=".cfg/.ini file to be converted",
    ),
    "output_file": dict(
        flags=("-o", "--output-file"),
        default="-",
        type=argparse.FileType("w"),
        help="file where to write the converted TOML (`stdout` by default)",
    ),
    "profile": dict(
        flags=("-p", "--profile"),
        default=None,
        help=f"a translation profile name, that will instruct {__package__} how "
        "to perform the most appropriate conversion. Available profiles:\n",
    ),
    "enable": dict(
        flags=("-E", "--enable"),
        nargs="+",
        dest="enable",
        metavar="TRANSFORMATION",
        help="enable one or more of the following processing options (optional):\n",
    ),
    "disable": dict(
        flags=("-D", "--disable"),
        nargs="+",
        dest="disable",
        default=(),
        metavar="TRANSFORMATION",
        help="disable one or more of the following processing options "
        "(active by default):\n",
    ),
    "verbose": dict(
        flags=("-v", "--verbose"),
        dest="loglevel",
        action="store_const",
        const=logging.INFO,
        help="set logging level to INFO",
    ),
    "very_verbose": dict(
        flags=("-vv", "--very-verbose"),
        dest="loglevel",
        action="store_const",
        const=logging.DEBUG,
        help="set logging level to DEBUG",
    ),
}


def __meta__(
    profiles: Sequence[Profile], augmentations: Sequence[ProfileAugmentation]
) -> Dict[str, dict]:
    """'Hyper parameters' to instruct :mod:`argparse` how to create the CLI"""
    meta = {k: v.copy() for k, v in META.items()}
    meta["profile"]["help"] += _choices_help(profiles, lambda x: x.help_text.strip())

    enable: List[ProfileAugmentation] = []
    disable: List[ProfileAugmentation] = []
    for aug in sorted(augmentations, key=lambda x: x.name):
        target = disable if aug.active_by_default else enable
        target.append(aug)

    for key, value in {"enable": enable, "disable": disable}.items():
        meta[key]["choices"] = [aug.name for aug in value]
        meta[key]["help"] += _choices_help(value)
        if not value:
            meta.pop(key, None)

    return meta


@critical_logging()
def parse_args(
    args: Sequence[str],
    profiles: Sequence[Profile],
    augmentations: Sequence[ProfileAugmentation],
) -> argparse.Namespace:
    """Parse command line parameters

    Args:
      args: command line parameters as list of strings (for example  ``["--help"]``).

    Returns: command line parameters namespace
    """
    description = "Automatically converts .cfg/.ini files into TOML"
    parser = argparse.ArgumentParser(description=description, formatter_class=Formatter)
    for opts in __meta__(profiles, augmentations).values():
        parser.add_argument(*opts.pop("flags", ()), **opts)
    parser.set_defaults(loglevel=logging.WARNING)
    params = parser.parse_args(args)
    profile_names = [p.name for p in profiles]
    profile = guess_profile(params.profile, params.input_file.name, profile_names)
    params.profile = profile
    opts = vars(params)
    active_augmentations = {k: True for k in (opts.get("enable") or ())}
    active_augmentations.update({k: False for k in (opts.get("disable") or ())})
    params.active_augmentations = active_augmentations
    return params


def setup_logging(loglevel: int):
    """Setup basic logging

    Args:
      loglevel: minimum loglevel for emitting messages
    """
    logformat = "[%(levelname)s] %(message)s"
    logging.basicConfig(level=loglevel, stream=sys.stderr, format=logformat)


@contextmanager
def exceptisons2exit():
    try:
        yield
    except Exception as ex:
        _logger.error(f"{ex.__class__.__name__}: {ex}")
        _logger.debug("Please check the following information:", exc_info=True)
        raise SystemExit(1)


@exceptisons2exit()
def run(args: Sequence[str] = ()):
    """Wrapper allowing :obj:`Translator` to be called in a CLI fashion.

    Instead of returning the value from :func:`Translator.translate`, it prints the
    result to the given ``output_file`` or ``stdout``.

    Args:
      args (List[str]): command line parameters as list of strings
          (for example  ``["--verbose", "setup.cfg"]``).
    """
    args = args or sys.argv[1:]
    translator = Translator()
    profiles = list(translator.profiles.values())
    profile_augmentations = list(translator.augmentations.values())
    params = parse_args(args, profiles, profile_augmentations)
    setup_logging(params.loglevel)
    out = translator.translate(
        params.input_file.read(), params.profile, params.active_augmentations
    )
    params.output_file.write(out)


class Formatter(argparse.RawTextHelpFormatter):
    # Since the stdlib does not specify what is the signature we need to implement in
    # order to create our own formatter, we are left no choice other then overwrite a
    # "private" method considered to be an implementation detail.

    def _split_lines(self, text, width):
        return list(chain.from_iterable(wrap(x, width) for x in text.splitlines()))


def guess_profile(profile: Optional[str], file_name: str, available: List[str]) -> str:
    if profile:
        return profile

    name = Path(file_name).name
    if name in available:
        # Optimize for the easy case
        _logger.info(f"Profile not explicitly set, {name!r} inferred.")
        return name

    fname = file_name.replace(pathsep, "/")
    for name in available:
        if fname.endswith(name):
            _logger.info(f"Profile not explicitly set, {name!r} inferred.")
            return name

    _logger.warning(f"Profile not explicitly set, using {DEFAULT_PROFILE!r}.")
    return DEFAULT_PROFILE


def _choices_help(choices: Sequence[CLIChoice], filt=lambda _: True) -> str:
    """``filt``: predicate function, only choices for which ``filt(c)`` is ``True`` will
    be included in the help text.
    """
    return "\n".join(_format_choice_help(c) for c in choices if filt(c))


def _flatten_str(text: str) -> str:
    if not text:
        return text
    text = " ".join(x.strip() for x in dedent(text).splitlines()).strip()
    text = text.rstrip(".,;").strip()
    return (text[0].lower() + text[1:]).strip()


def _format_choice_help(choice: CLIChoice) -> str:
    return f'- "{choice.name}": {_flatten_str(choice.help_text)}.'



# https://mypy.readthedocs.io/en/stable/config_file.html
import string
from collections.abc import MutableMapping, MutableSequence
from functools import partial
from typing import List, TypeVar, cast

from ..transformations import coerce_scalar, split_list
from ..types import IntermediateRepr
from ..types import Transformation as T
from ..types import Translator

M = TypeVar("M", bound=MutableMapping)
R = TypeVar("R", bound=IntermediateRepr)

list_comma = partial(split_list, sep=",")


def activate(translator: Translator):
    plugin = Mypy()
    for file in ("setup.cfg", "mypy.ini", ".mypy.ini"):
        translator[file].intermediate_processors.append(plugin.process_values)

    translator["mypy.ini"].help_text = plugin.__doc__ or ""


class Mypy:
    """Convert settings to 'pyproject.toml' equivalent"""

    LIST_VALUES = (
        "files",
        "always_false",
        "disable_error_code",
        "plugins",
    )

    DONT_TOUCH = ("python_version",)

    def process_values(self, doc: M) -> M:
        for parent in (doc, doc.get("tool", {})):
            for key in list(parent.keys()):  # need to be eager: we will delete elements
                key_ = key[-1] if isinstance(key, tuple) else key
                if not isinstance(key_, str):
                    continue
                name = key_.strip('"' + string.whitespace)
                if name.startswith("mypy-"):
                    overrides = self.get_or_create_overrides(parent)
                    self.process_overrides(parent.pop(key), overrides, name)
                elif name == "mypy":
                    self.process_options(parent[key])
        return doc

    def process_overrides(self, section: R, overrides: MutableSequence, name: str) -> R:
        section = self.process_options(section)
        modules = [n.replace("mypy-", "") for n in name.split(",")]
        self.add_overrided_modules(section, name, modules)
        overrides.append(section)
        return section

    def process_options(self, section: M) -> M:
        for field in section:
            if field in self.DONT_TOUCH:
                continue
            fn: T = split_list if field in self.LIST_VALUES else coerce_scalar
            section[field] = fn(section[field])
        return section

    def get_or_create_overrides(self, parent: MutableMapping) -> MutableSequence:
        mypy = parent.setdefault("mypy", IntermediateRepr())
        return cast(MutableSequence, mypy.setdefault("overrides", []))

    def add_overrided_modules(self, section: R, name: str, modules: List[str]):
        if not isinstance(section, IntermediateRepr):
            raise ValueError(f"Expecting section {name} to be an IntermediateRepr")
        if "module" not in section:
            section.insert(0, "module", modules)



"""Profile-independent tasks implemented via *profile augmentation*."""
import re
from functools import wraps
from typing import Callable

from ..types import Profile, Translator

NEWLINES = re.compile(r"\n+", re.M)
TABLE_START = re.compile(r"^\[(.*)\]", re.M)
EMPTY_TABLES = re.compile(r"^\[(.*)\]\n+\[(\1\.(?:.*))\]", re.M)
TRAILING_SPACES = re.compile(r"[ \t]+$", re.M)


def activate(translator: Translator):
    tasks = [remove_trailing_spaces, normalise_newlines, remove_empty_table_headers]
    for task in tasks:
        translator.augment_profiles(post_process(task), active_by_default=True)


def post_process(fn: Callable[[str], str]):
    @wraps(fn)
    def _augmentation(profile: Profile):
        profile.post_processors.append(fn)

    return _augmentation


def remove_trailing_spaces(text: str) -> str:
    """Remove trailing spaces"""
    return TRAILING_SPACES.sub("", text)


def normalise_newlines(text: str) -> str:
    """Make sure every table is preceded by an empty newline, but remove them elsewhere
    in the output TOML document.
    """
    text = NEWLINES.sub(r"\n", text)
    return TABLE_START.sub(r"\n[\1]", text)


def remove_empty_table_headers(text: str) -> str:
    """Remove empty TOML table headers"""
    return EMPTY_TABLES.sub(r"[\2]", text).strip()



from textwrap import dedent

import pytest

from ini2toml.drivers.full_toml import loads
from ini2toml.plugins import best_effort
from ini2toml.translator import Translator


@pytest.fixture
def translate():
    translator = Translator(plugins=[best_effort.activate])
    return lambda text: translator.translate(text, "best_effort")


@pytest.fixture
def ini2tomlobj(translate):
    return lambda text: loads(translate(text))


def test_best_effort(ini2tomlobj):
    example = """\
    [section]
    a = 1
    b = 0.42
    c = false
    d = on
    f =
        a
        b
    e =
        a=1
        b=2
    [nested.section]
    value = string
    """
    doc = ini2tomlobj(dedent(example))
    assert doc["section"]["a"] == 1
    assert doc["section"]["b"] == 0.42
    assert doc["section"]["c"] is False
    assert doc["section"]["d"] is True
    assert doc["section"]["f"] == ["a", "b"]
    assert doc["section"]["e"] == {"a": "1", "b": "2"}
    assert doc["nested"]["section"]["value"] == "string"



# The code in this module is mostly borrowed/adapted from PyScaffold and was originally
# published under the MIT license
# The original PyScaffold license can be found in 'tests/examples/pyscaffold'

import sys
from textwrap import dedent
from typing import Callable, Iterable, List, Optional

from .. import __version__
from ..types import Plugin

if sys.version_info[:2] >= (3, 8):  # pragma: no cover
    # TODO: Import directly (no need for conditional) when `python_requires = >= 3.8`
    from importlib.metadata import EntryPoint, entry_points
else:  # pragma: no cover
    from importlib_metadata import EntryPoint, entry_points


ENTRYPOINT_GROUP = "ini2toml.processing"


def iterate_entry_points(group=ENTRYPOINT_GROUP) -> Iterable[EntryPoint]:
    """Produces a generator yielding an EntryPoint object for each plugin registered
    via `setuptools`_ entry point mechanism.

    This method can be used in conjunction with :obj:`load_from_entry_point` to filter
    the plugins before actually loading them.


    .. _setuptools: https://setuptools.pypa.io/en/latest/userguide/entry_point.html
    """  # noqa
    entries = entry_points()
    if hasattr(entries, "select"):
        # The select method was introduced in importlib_metadata 3.9 (and Python 3.10)
        # and the previous dict interface was declared deprecated
        entries_ = entries.select(group=group)  # type: ignore
    else:
        # TODO: Once Python 3.10 becomes the oldest version supported, this fallback and
        #       conditional statement can be removed.
        entries_ = (plugin for plugin in entries.get(group, []))
    return sorted(entries_, key=lambda e: e.name)


def load_from_entry_point(entry_point: EntryPoint) -> Plugin:
    """Carefully load the plugin, raising a meaningful message in case of errors"""
    try:
        return entry_point.load()
    except Exception as ex:
        raise ErrorLoadingPlugin(entry_point=entry_point) from ex


def list_from_entry_points(
    group: str = ENTRYPOINT_GROUP,
    filtering: Callable[[EntryPoint], bool] = lambda _: True,
) -> List[Plugin]:
    """Produces a list of plugin objects for each plugin registered
    via `setuptools`_ entry point mechanism.

    Args:
        group: name of the setuptools' entry_point group where plugins is being
            registered
        filtering: function returning a boolean deciding if the entry point should be
            loaded and included (or not) in the final list. A ``True`` return means the
            plugin should be included.

    .. _setuptools: https://setuptools.pypa.io/en/latest/userguide/entry_point.html
    """  # noqa
    return [
        load_from_entry_point(e) for e in iterate_entry_points(group) if filtering(e)
    ]


class ErrorLoadingPlugin(RuntimeError):
    """There was an error loading '{plugin}'.
    Please make sure you have installed a version of the plugin that is compatible
    with {package} {version}. You can also try uninstalling it.
    """

    def __init__(self, plugin: str = "", entry_point: Optional[EntryPoint] = None):
        if entry_point and not plugin:
            plugin = getattr(entry_point, "module", entry_point.name)

        sub = dict(package=__package__, version=__version__, plugin=plugin)
        msg = dedent(self.__doc__ or "").format(**sub).splitlines()
        super().__init__(f"{msg[0]}\n{' '.join(msg[1:])}")



import logging
import re
from functools import partial, reduce
from itertools import chain
from typing import Dict, List, Mapping, Sequence, Set, Tuple, Type, TypeVar, Union, cast

from ..transformations import (
    apply,
    coerce_bool,
    deprecated,
    kebab_case,
    remove_prefixes,
    split_comment,
    split_kv_pairs,
    split_list,
)
from ..types import CommentKey, HiddenKey
from ..types import IntermediateRepr as IR
from ..types import Transformation, Translator, WhitespaceKey
from .best_effort import BestEffort

try:
    from setuptools._distutils import command as distutils_commands
except ImportError:  # pragma: no cover
    from distutils import command as distutils_commands

R = TypeVar("R", bound=IR)

RenameRules = Dict[Tuple[str, ...], Union[Tuple[Union[str, int], ...], None]]
ProcessingRules = Dict[Tuple[str, ...], Transformation]


_logger = logging.getLogger(__name__)

chain_iter = chain.from_iterable

# Functions that split values from comments and parse those values
split_directive = partial(split_kv_pairs, key_sep=":")
split_list_comma = partial(split_list, sep=",", subsplit_dangling=False)
split_list_semi = partial(split_list, sep=";", subsplit_dangling=False)
split_hash_comment = partial(split_comment, comment_prefixes="#")  # avoid splitting `;`
split_bool = partial(split_comment, coerce_fn=coerce_bool)
split_kv_nocomments = partial(split_kv_pairs, comment_prefixes="")
split_kv_of_lists = partial(split_kv_pairs, coerce_fn=split_list_comma)

SECTION_SPLITTER = re.compile(r"\.|:")
SETUPTOOLS_SECTIONS = ("metadata", "options")
SKIP_CHILD_NORMALISATION = ("options.entry_points",)
COMMAND_SECTIONS = (
    "global",
    "alias",
    "install",
    "develop",
    "sdist",
    "bdist",
    "bdist_wheel",
    *getattr(distutils_commands, "__all__", []),
)


def activate(translator: Translator):
    plugin = SetuptoolsPEP621()
    profile = translator["setup.cfg"]
    profile.intermediate_processors += [plugin.normalise_keys, plugin.pep621_transform]
    profile.help_text = plugin.__doc__ or ""


class SetuptoolsPEP621:
    """Convert settings to 'pyproject.toml' based on PEP 621"""

    BUILD_REQUIRES = ("setuptools", "wheel")

    def __init__(self):
        self._be = BestEffort(key_sep="=")

    @classmethod
    def template(
        cls,
        ir_cls: Type[R] = IR,  # type: ignore
        build_requires: Sequence[str] = (),
    ) -> R:
        build_system = {
            "requires": [*(build_requires or cls.BUILD_REQUIRES)],
            # ^ NOTE: the code ahead assumes no version
            "build-backend": "setuptools.build_meta",
        }
        tpl = {
            "metadata": ir_cls(),  # NOTE: will be renamed later
            "build-system": ir_cls(build_system),  # type: ignore
            "tool": ir_cls(),
        }
        return ir_cls(tpl)  # type: ignore

    def setupcfg_aliases(self):
        """``setup.cfg`` aliases as defined in:
        https://setuptools.pypa.io/en/stable/userguide/declarative_config.html
        """
        return {
            "classifier": "classifiers",
            "summary": "description",
            "platform": "platforms",
            "license-file": "license-files",
            "home-page": "url",
        }

    def setupcfg_directives(self):
        """``setup.cfg`` directives, as defined in:
        https://setuptools.pypa.io/en/stable/userguide/declarative_config.html
        """
        return {
            ("metadata", "version"): ("file", "attr"),
            ("metadata", "classifiers"): ("file",),
            ("metadata", "description"): ("file",),
            ("metadata", "long-description"): ("file",),
            ("options", "entry-points"): ("file",),
            ("options", "packages"): ("find", "find_namespace"),
        }

    def processing_rules(self) -> ProcessingRules:
        """Value type processing, as defined in:
        https://setuptools.pypa.io/en/stable/userguide/declarative_config.html
        """
        return {
            ("metadata", "classifiers"): split_list_comma,
            # => ("metadata", "license_files",): in PEP621 should be a single file
            ("metadata", "keywords"): split_list_comma,
            # => ("metadata", "project-urls") => merge_and_rename_urls
            # ("metadata", "long-description-content-type") =>
            #     merge_and_rename_long_description_and_content_type
            # ---- the following options are originally part of `metadata`,
            #      but we move it to "options" because they are not part of PEP 621
            ("options", "provides"): split_list_comma,
            ("options", "requires"): split_list_comma,
            ("options", "obsoletes"): split_list_comma,
            ("options", "platforms"): split_list_comma,
            # ----
            ("options", "zip-safe"): split_bool,
            ("options", "setup-requires"): split_list_semi,
            ("options", "install-requires"): split_list_semi,
            ("options", "tests-require"): split_list_semi,
            ("options", "script-files"): split_list_comma,
            # ^ --> We rename scripts to script-files
            ("options", "eager-resources"): split_list_comma,
            ("options", "dependency-links"): split_list_comma,
            ("options", "include-package-data"): split_bool,
            ("options", "packages"): split_list_comma,
            ("options", "package-dir"): split_kv_pairs,
            ("options", "namespace-packages"): split_list_comma,
            ("options", "py-modules"): split_list_comma,
            ("options", "data-files"): deprecated("data-files", split_kv_of_lists),
            ("options.packages.find", "include"): split_list_comma,
            ("options.packages.find", "exclude"): split_list_comma,
            ("options.packages.find-namespace", "include"): split_list_comma,
            ("options.packages.find-namespace", "exclude"): split_list_comma,
        }
        # See also dependent_processing_rules
        # Everything else should use split_comment

    def dependent_processing_rules(self, doc: IR) -> ProcessingRules:
        """Dynamically create processing rules, such as :func:`value_processing` based
        on the existing document.
        """
        groups: Mapping[str, Transformation] = {
            "options.extras-require": split_list_semi,
            "options.package-data": split_list_comma,
            "options.exclude-package-data": split_list_comma,
            "options.data-files": split_list_comma,
            # ----
            # "options.entry-points" => moved to project
            # console-scripts and gui-scripts already handled in
            # move_and_split_entrypoints
            "project:entry-points": split_kv_pairs,
        }
        return {
            (g, k): fn
            for g, fn in groups.items()
            for k in doc.get(g, ())
            if isinstance(k, str)
        }

    def merge_and_rename_urls(self, doc: R) -> R:
        #  url => urls.homepage
        #  download-url => urls.download
        #  project-urls => urls
        metadata = cast(IR, doc["metadata"])
        new_urls = (
            f"{dest} = {metadata.pop(orig)}"
            for orig, dest in [("url", "Homepage"), ("download-url", "Download")]
            if orig in metadata
        )
        urls = "\n".join(chain(new_urls, [metadata.get("project-urls", "")]))
        urls = urls.strip()
        if urls:
            urls_kv = split_kv_nocomments(urls)
            keys = ("project-urls", "url", "download-url")
            metadata.replace_first_remove_others(keys, "urls", urls_kv)
        return doc

    def merge_authors_maintainers_and_emails(self, doc: R) -> R:
        # author OR maintainer => author.name
        # author-email OR maintainer-email => author.email
        metadata: IR = doc["metadata"]
        author_ = split_comment(metadata.get("author", ""))
        maintainer_ = split_comment(metadata.get("maintainer", ""))
        names_ = (author_, maintainer_)
        names = chain_iter(n.value_or("").strip().split(",") for n in names_)
        a_email_ = split_comment(metadata.get("author-email", ""))
        m_email_ = split_comment(metadata.get("maintainer-email", ""))
        emails_ = (a_email_, m_email_)
        emails = chain_iter(n.value_or("").strip().split(",") for n in emails_)
        comments = [o.comment for o in chain(names_, emails_) if o.has_comment()]

        combined_ = {e: n for n, e in zip(names, emails) if n}  # deduplicate
        out = [{"name": n, "email": e} for e, n in combined_.items()]
        if out:
            keys = ("author", "maintainer", "author-email", "maintainer-email")
            i = metadata.replace_first_remove_others(keys, "author", out)
            for j, cmt in enumerate(comments):
                metadata.insert(j + i + 1, CommentKey(), cmt)
        return doc

    def merge_and_rename_long_description_and_content_type(self, doc: R) -> R:
        # long_description.file => readme.file
        # long_description => readme.text
        # long-description-content-type => readme.content-type
        metadata: IR = doc["metadata"]
        if "long-description" not in metadata:
            metadata.pop("long-description-content-type", None)
            return doc
        long_desc = metadata["long-description"].strip()
        readme: dict = {}
        if long_desc.startswith("file:"):
            readme = {"file": remove_prefixes(long_desc, ("file:",)).strip()}
        elif long_desc:
            readme = {"text": long_desc}
        content_type = metadata.pop("long-description-content-type", None)
        if content_type:
            readme["content-type"] = split_hash_comment(content_type)
        if len(list(readme.keys())) == 1 and "file" in readme:
            metadata["long-description"] = split_comment(readme["file"])
        else:
            readme_ = IR({k: split_comment(v) for k, v in readme.items()})
            metadata["long-description"] = readme_
        metadata.rename("long-description", "readme")
        return doc

    def merge_license_and_files(self, doc: R) -> R:
        # Prepare license before pep621_renaming
        # license-files => license.file
        # license => license.text
        metadata: IR = doc["metadata"]
        naming = {"license-files": "file", "license": "text"}
        items = [
            (v, split_comment(metadata[k])) for k, v in naming.items() if k in metadata
        ]
        if not metadata or not items:
            return doc
        # 'file' and 'text' are mutually exclusive in PEP 621
        license = IR(dict(items[:1]))
        metadata.replace_first_remove_others(list(naming.keys()), "license", license)
        return doc

    def move_and_split_entrypoints(self, doc: R) -> R:
        # This is part of pep621_renaming
        # "entry-points"."console-scripts" => "scripts"
        # "entry-points"."gui-scripts" => "gui-scripts"
        entrypoints: IR = doc.get("options.entry-points", IR())
        if not entrypoints:
            return doc
        doc.rename("options.entry-points", "project:entry-points")
        # ^ use `:` to guarantee it is split later
        script_keys = ["console-scripts", "gui-scripts"]
        script_keys += [k.replace("-", "_") for k in script_keys]
        keys = (k for k in script_keys if k in entrypoints)
        for key in keys:
            scripts = split_kv_pairs(entrypoints.pop(key)).to_ir()
            new_key = key.replace("_", "-").replace("console-", "")
            doc.append(f"project:{new_key}", scripts)
        if not entrypoints or all(isinstance(k, WhitespaceKey) for k in entrypoints):
            doc.pop("project:entry-points")
        return doc

    def move_options_missing_in_pep621(self, doc: R) -> R:
        # ---- Things in "options" that are covered by PEP 621 ----
        # First we handle simple options
        naming = {
            "python-requires": "requires-python",
            "install-requires": "dependencies",
            "entry-points": "entry-points",
        }
        metadata, options = doc["metadata"], doc["options"]
        metadata.update({v: options.pop(k) for k, v in naming.items() if k in options})

        # Then we handle entire sections:
        naming = {"extras-require": "optional-dependencies"}
        for src, target in naming.items():
            doc.rename(f"options.{src}", f"project:{target}", ignore_missing=True)
        return doc

    def rename_script_files(self, doc: R) -> R:
        # setuptools define a ``options.scripts`` parameters that refer to
        # script files, not created via enty-points
        # To avoid confution with PEP621 scripts (generated via entry-points)
        # let's rename this field to `script-files`
        doc["options"].rename("scripts", "script-files", ignore_missing=True)
        return doc

    def remove_metadata_not_in_pep621(self, doc: R) -> R:
        # ---- setuptools metadata without correspondence in PEP 621 ----
        specific = ["platforms", "provides", "obsoletes"]
        metadata, options = doc["metadata"], doc["options"]
        options.update({k: metadata.pop(k) for k in specific if k in metadata})
        return doc

    def parse_setup_py_command_options(self, doc: R) -> R:
        # ---- distutils/setuptools command specifics outside of "options" ----
        sections = list(doc.keys())
        commands = _distutils_commands()
        for k in sections:
            if isinstance(k, str) and k in commands:
                section = self._be.apply_best_effort_to_section(doc[k])
                for option in section:
                    if isinstance(option, str):
                        section.rename(option, self.normalise_key(option))
                doc[k] = section
                doc.rename(k, ("distutils", k))
        return doc

    def convert_directives(self, out: R) -> R:
        for (section, option), directives in self.setupcfg_directives().items():
            value = out.get(section, {}).get(option, None)
            if isinstance(value, str) and any(
                value.strip().startswith(f"{d}:") for d in directives
            ):
                out[section][option] = split_directive(value)
        return out

    def split_subtables(self, out: R) -> R:
        """Setuptools emulate nested sections (e.g.: ``options.extras_require``)"""
        sections = [
            k
            for k in out.keys()
            if isinstance(k, str) and (k.startswith("options.") or ":" in k)
        ]
        for section in sections:
            new_key = SECTION_SPLITTER.split(section)
            if section.startswith("options."):
                new_key = ["tool", "setuptools", *new_key[1:]]
            out.rename(section, tuple(new_key))
        return out

    def apply_value_processing(self, doc: R) -> R:
        default = {
            (name, option): split_comment
            for name, section in doc.items()
            if name in ("metadata", "options")
            for option in section
            if isinstance(option, (str, tuple))
        }
        transformations: dict = {
            **default,
            **self.processing_rules(),
            **self.dependent_processing_rules(doc),
        }
        for (section, option), fn in transformations.items():
            value = doc.get(section, {}).get(option, None)
            if value is not None:
                doc[section][option] = fn(value)
        return doc

    def fix_dynamic(self, doc: R) -> R:
        potential = ["version", "classifiers", "description"]
        metadata, options = doc["metadata"], doc["options"]

        fields = [f for f in potential if isdirective(metadata.get(f, None))]
        dynamic = {f: split_directive(metadata.pop(f, None)) for f in fields}
        if "version" not in metadata and "version" not in dynamic:
            msg = (
                "No `version` was found in `[metadata]`, `ini2toml` will assume it is "
                "defined by tools like `setuptools-scm` or in `setup.py`. "
                "Automatically adding it to `dynamic` (in accordance with PEP 621)"
            )
            _logger.debug(msg)
            fields.insert(0, "version")

        extras: List[str] = []
        ep = options.pop("entry-points", None)
        if isdirective(ep, valid=("file",)):
            fields.append("entry-points")
            dynamic["entry-points"] = split_directive(ep)
            extras = ["scripts", "gui-scripts"]
        if not fields:
            return doc
        metadata.setdefault("dynamic", []).extend(fields + extras)

        if dynamic:
            doc.setdefault("options.dynamic", IR()).update(dynamic)
            # ^ later `options.dynamic` is converted to `tool.setuptools.dynamic`
        return doc

    def fix_packages(self, doc: R) -> R:
        options = doc["options"]
        packages = options.get("packages", "").strip()
        if not packages:
            return doc
        prefixes = ["find", *[f"find{_}namespace" for _ in "_-"]]
        prefix = next((p for p in prefixes if packages.startswith(f"{p}:")), None)
        if not prefix:
            return doc
        kebab_prefix = prefix.replace("_", "-")
        options["packages"] = {kebab_prefix: {}}
        if "options.packages.find" in doc:
            options.pop("packages")
            doc.rename("options.packages.find", f"options.packages.{kebab_prefix}")
        return doc

    def fix_setup_requires(self, doc: R) -> R:
        """Add mandatory dependencies if they are missing"""
        options = doc["options"]
        requirements = options.get("setup-requires", "")
        if not requirements:
            return doc
        req = requirements.splitlines()
        if len(req) > 1:
            joiner = "\n"
        else:
            joiner = "; "
            req = req[0].split(";")
        build_req = doc["build-system"]["requires"]
        req.extend(d for d in build_req if d not in requirements)
        options["setup-requires"] = joiner.join(req)
        return doc

    def move_setup_requires(self, doc: R) -> R:
        options = doc["options"]
        if "setup-requires" in options:
            doc["build-system"]["requires"] = options.pop("setup-requires")
        return doc

    def ensure_pep518(self, doc: R) -> R:
        """PEP 518 specifies that any other tool adding configuration under
        ``pyproject.toml`` should use the ``tool`` table. This means that the only
        top-level keys are ``build-system``, ``project`` and ``tool``
        """
        allowed = ("build-system", "project", "tool", "metadata", "options")
        allowed_prefixes = ("options.", "project:")
        for k in list(doc.keys()):
            key = k
            rest: Sequence = ()
            if isinstance(k, tuple):
                key, *rest = k
            if isinstance(key, HiddenKey):
                continue
            if not (key in allowed or any(key.startswith(p) for p in allowed_prefixes)):
                doc.rename(k, ("tool", key, *rest))
        return doc

    def pep621_transform(self, doc: R) -> R:
        """Rules are applied sequentially and therefore can interfere with the following
        ones. Please notice that renaming is applied after value processing.
        """
        transformations = [
            # --- transformations mainly focusing on PEP 621 ---
            self.merge_and_rename_urls,
            self.merge_authors_maintainers_and_emails,
            self.merge_license_and_files,
            self.merge_and_rename_long_description_and_content_type,
            self.move_and_split_entrypoints,
            # --- General fixes
            self.rename_script_files,
            self.remove_metadata_not_in_pep621,
            self.fix_packages,
            self.fix_setup_requires,
            self.fix_dynamic,
            # --- value processing and type changes ---
            self.convert_directives,
            self.apply_value_processing,
            self.parse_setup_py_command_options,
            # --- steps that depend on the values being processed ---
            self.move_setup_requires,
            self.move_options_missing_in_pep621,
            # --- final steps ---
            self.split_subtables,
            self.ensure_pep518,
        ]
        out = self.template(doc.__class__)
        out.update(doc)
        out.setdefault("metadata", IR())
        out.setdefault("options", IR())
        out = reduce(apply, transformations, out)
        out.rename("metadata", "project", ignore_missing=True)
        out.rename("options", ("tool", "setuptools"), ignore_missing=True)
        return out

    def normalise_keys(self, cfg: R) -> R:
        """Normalise keys in ``setup.cfg``, by replacing aliases with cannonic names
        and replacing the snake case with kebab case.

        .. note:: Although setuptools recently deprecated kebab case in ``setup.cfg``
           ``pyproject.toml`` seems to use it as a convention, so this normalisation
           makes more sense for the translation.
        """
        # Normalise for the same convention as pyproject
        for i in range(len(cfg.order)):
            section_name = cfg.order[i]
            if not isinstance(section_name, str):
                continue
            if not any(section_name.startswith(s) for s in SETUPTOOLS_SECTIONS):
                continue
            section = cfg[section_name]
            cfg.rename(section_name, kebab_case(section_name))
            if any(section_name.startswith(s) for s in SKIP_CHILD_NORMALISATION):
                continue
            for j in range(len(section.order)):
                option_name = section.order[j]
                if not isinstance(option_name, str):
                    continue
                section.rename(option_name, self.normalise_key(option_name))
        # Normalise aliases
        metadata = cfg.get("metadata")
        if not metadata:
            return cfg
        for alias, cannonic in self.setupcfg_aliases().items():
            if alias in metadata:
                metadata.rename(alias, cannonic)
        return cfg

    def normalise_key(self, key: str) -> str:
        """Normalise a single key for option"""
        return kebab_case(key)


# ---- Helpers ----


def isdirective(value, valid=("file", "attr")) -> bool:
    return isinstance(value, str) and any(value.startswith(f"{p}:") for p in valid)


def _distutils_commands() -> Set[str]:
    try:
        from . import iterate_entry_points

        commands = [ep.name for ep in iterate_entry_points("distutils.commands")]
    except Exception:
        commands = []
    return {*commands, *COMMAND_SECTIONS}
